#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>

#include "fits2lc.h"

static double Xcenter, Ycenter, Zcenter, posvink, pixmm;
static double phas, cosvink, sinvink, cosph2, sinph2, cosj_dec, sinj_dec, cosj_RA, sinj_RA;

int getRAdec_init(int jemx, int aux_orbit, int aux_pid, 
   	double p_RA[10000], double p_dec[10000], double p_posang[10000])
/*   Scan through the pointing data to find the pointing direction corresponding to aux_orbit/aux_pid */
{
   int i, n, p_orbit, p_pid, p_type, p_year, p_month, p_day, p_hour, p_min, p_sec, pointval;
   char dummy[100], pointfil[200];
   float p_durat, indat[2];
   FILE *pofil;

     sprintf(pointfil, "%s/reference/alignment.dat", sweethome);
     pofil = fopen(pointfil, "rt");
     if (pofil == NULL) {
         printf("No alignment file found! (%s)\n", pointfil);
         exit(0);
     } else {
         for (i=0; i<200; i++) {
            fgets(dummy, 100, pofil);
            if (feof(pofil)) goto q_end;
            sscanf(dummy, "%d", &n);
            if (n == 4)   {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); coll_r = indat[jemx];}
            if (n == 5)   {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); mask_r = indat[jemx];}
            if (n == 6)   {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); mask_height = indat[jemx];}

            if (n == 21)  {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); FOV_radius = indat[jemx];}

            if (n == 100) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); scaleX = indat[jemx];}
            if (n == 101) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); scaleY = indat[jemx];}
            if (n == 102) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); phase2 = indat[jemx];}
            if (n == 103) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); x2x = indat[jemx];}
            if (n == 104) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); y2y = indat[jemx];}
            if (n == 105) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offX = indat[jemx];}
            if (n == 106) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offY = indat[jemx];}
            if (n == 107) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offQ = indat[jemx];}
            if (n == 108) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offR = indat[jemx];}
         }
      }
q_end:

      pixmm = (coll_r + mask_r) / FOV_radius;
      pix2rad = pixmm / mask_height;
      offQ *= pix2rad;
      offR *= pix2rad;
      
/*      printf("Read scale parameters: %lf %lf %lf %lf %lf %lf %lf %lf %lf, derived pix2rad: %10.8lf\n", 
         scaleX, scaleY, phase2, x2x, y2y, offX, offY, offQ, offR, pix2rad);
*/
      fclose(pofil);

      pointval = 0;
      sprintf(pointfil, "%s/pointing.dat", sweethome2);
      pofil = fopen(pointfil, "rt");
      if (pofil == NULL) {
         printf("No pointing file found! (%s)\n", pointfil);
      } else {
         for (i=0; i<1000000; i++) {
            fgets(dummy, 100, pofil);
            if (feof(pofil)) goto p_end;
            sscanf(dummy, "%d %d %d %lf %lf %lf %d %d %d %d %d %d %f",
               &p_orbit, &p_pid, &p_type, &p_RA[aux_pid], &p_dec[aux_pid], &p_posang[aux_pid], 
               &p_year, &p_month, &p_day, &p_hour, &p_min, &p_sec, &p_durat);
            if ((p_orbit == aux_orbit) && (p_pid == aux_pid) && ((p_type % 2) == 0)) {
               pointval = 1;
               goto p_end;
            }
         }
      }
p_end:
      fclose(pofil);
   
      if (pointval > 0) {
/*       printf("Pointing: RA: %8.4lf,  dec: %8.4lf,  posang: %8.4lf\n", p_RA, p_dec, p_posang);
*/
      	 j_RA = p_RA[aux_pid] * gtorad;
         j_dec = p_dec[aux_pid] * gtorad;
      
         Xcenter = cos(j_dec) * cos(j_RA);		/* X towards spring equinox */
         Ycenter = cos(j_dec) * sin(j_RA);		/* Y perpendicular to X and Z */
         Zcenter = sin(j_dec);			/* Z towards North */

      } else printf("No matching pointing data found for orbit %3d and pid: %3d\n", aux_orbit, aux_pid);

      posvink = (180.0 - p_posang[aux_pid]) * gtorad;

      cosvink = cos(posvink);
      sinvink = sin(posvink);
      cosph2 = cos(phase2);
      sinph2 = sin(phase2);
      cosj_dec = cos(j_dec);
      sinj_dec = sin(j_dec);
      cosj_RA = cos(j_RA);
      sinj_RA = sin(j_RA);
      
/*    printf("Pointing transformation initialized\n");
*/
      return(pointval);
}

/* *********************************************************** */

int getRAdec(float x, float y, int jemx, float *s_RA, float *s_dec)
{
   
   double srcQ, srcR, Xcorr, Ycorr, SRCY, SRCZ, SRCX, SRC_R, Xsrc, Ysrc, Zsrc, Rsrc;
   float avX, avY, avr;
               
   avX = x - (float)(FOV_radius);
   avY = y - (float)(FOV_radius);
   avX = avX + x2x * avX * avX;
   avY = avY + y2y * avY * avY;
      
   avr = sqrt((double)(avX*avX + avY*avY));

   Xcorr = (avX + offX) * cosph2 - (avY + offY) * sinph2;
   Ycorr = (avY + offY) * cosph2 + (avX + offX) * sinph2;

   Xcorr = -Xcorr;
   Ycorr = -Ycorr;
            
   srcQ = (Xcorr * scaleX * cosvink + Ycorr * scaleY * sinvink) * pix2rad + offQ; 
   srcR = (-Xcorr * scaleX * sinvink + Ycorr * scaleY * cosvink) * pix2rad + offR;	
               /* the Q-axis is in the mask plane, parallel to equator, srcQ is in radians */
               /* the R-coordinate is in mask plane, perpendicular to Q, srcR is in radians */

   Zsrc = srcR * cosj_dec;
   Xsrc = -srcQ * sinj_RA - srcR * sinj_dec * cosj_RA;
   Ysrc =  srcQ * cosj_RA - srcR * sinj_dec * sinj_RA;
   
   SRCX = Xcenter + Xsrc;
   SRCY = Ycenter + Ysrc;
   SRCZ = Zcenter + Zsrc;

   SRC_R = sqrt(SRCX*SRCX + SRCY*SRCY + SRCZ*SRCZ);       
                 
   *s_dec = asin(SRCZ / SRC_R);
   *s_RA = atan2(SRCY, SRCX);
   
   if (*s_RA < 0.0) *s_RA += twopi;                

   *s_dec /= gtorad;
   *s_RA /= gtorad; 
   
   return(0);
}

/* *********************************************************** */

int getxy(float s_RA, float s_dec, int jemx, float *x, float *y)
{
   
   double srcQ, srcR, Xcorr, Ycorr, SRCY, SRCZ, SRCX, SRC_R, Xsrc, Ysrc, Zsrc, Rsrc;
   double dist2, dist, ksi, cosksi, rr2;
   float avX, avY, avr;
               
   s_dec *= gtorad;
   s_RA *= gtorad; 
   
   SRCX = cos((double)(s_dec)) * cos((double)(s_RA));
   SRCY = cos((double)(s_dec)) * sin((double)(s_RA));
   SRCZ = sin((double)(s_dec));   

   dist2 = (SRCX-Xcenter)*(SRCX-Xcenter) + (SRCY-Ycenter)*(SRCY-Ycenter) + (SRCZ-Zcenter)*(SRCZ-Zcenter);
   dist = sqrt(dist2);
   
   ksi = atan(dist / sqrt(4.0 - dist2));
   
   cosksi = cos(2.0 * ksi);
   SRCX /= cosksi;
   SRCY /= cosksi;
   SRCZ /= cosksi;
   
   Xsrc = SRCX  - Xcenter;
   Ysrc = SRCY  - Ycenter;
   Zsrc = SRCZ  - Zcenter;
   
   srcR = Zsrc / cosj_dec;
   rr2 = Xsrc*Xsrc + Ysrc*Ysrc + Zsrc*Zsrc;
   srcQ = sqrt(rr2 - srcR * srcR);
   if (j_RA > s_RA) srcQ = -srcQ;
   
   Xcorr = ((srcQ - offQ) * cosvink - (srcR - offR) * sinvink) / pix2rad;
   Ycorr = ((srcQ - offQ) * sinvink + (srcR - offR) * cosvink) / pix2rad;

   Xcorr = -Xcorr;
   Ycorr = -Ycorr;
            
   avX = (Xcorr * cosph2 + Ycorr * sinph2) - offX;
   avY = (-Xcorr * sinph2 + Ycorr * cosph2) - offY;
   
   *x = avX - x2x * avX * avX;
   *y = avY - y2y * avY * avY;
   
   *x += (float)(FOV_radius);
   *y += (float)(FOV_radius);

   return(0);

}

