#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>

#include "midi_b.h"

int irosloop(float shadow[detdim], int jemx, int ee, float xx[sourcmax], float yy[sourcmax], float norm_d[sourcmax], 
   	double solut[sourcmax], float sumfun[sourcmax], float *goodrms, double *sh_rms2, int *goodfits, 
        int *goodnsrc, float resid[detdim], float edge, float goodbet_g[sourcmax], int iter, double limit_fraction)
{
      int i, j, k, m, sky_index, blindnum, tours, updates[sourcmax], nsrc, ngrp, fitcount;
      int mini_rms, mini_const, sxy, sxy_list[sourcmax], blindspots[sourcmax], just_updated;
      int smx[max_candidates], smy[max_candidates];
      float valmin_rms, valmin_const, xnow, ynow, varim[skydim], shadow_counts;
      double amx[max_candidates+1], rms;
      char filnavn[100];


      blindnum = 0;
      tours = 0;
      for (i=0; i<sourcmax; i++) updates[i] = 0; 

      nsrc = backmodels;
      ngrp = nsrc - 1;

      fun(-1, 1, ee, xx, yy, jemx, norm_d, 0); 	/* reset fitting functions */

      /* ***************************************************************** */
      /*                                                                   */
      /*      Backproject raw shadowgram to form variance image            */
      /*                                                                   */
      /* ***************************************************************** */
      
      shadow_counts = backpro_d(shadow, varim);

      for (i=0; i<skydim; i++) fitsimage[ee][0][i] = varim[i] * 1.0555; 
         	/* 1.0555 = 1 + 1/18 (background contribution to variance) */

      /* ******************************************************** */
      /*                                                          */
      /*       Fit shadowgram with background components          */
      /*                                                          */
      /* ******************************************************** */
      
	/* imagefit returns a solution to the fit process in "solut" and a residual shadowgram in "resid" */

      rms = imagefit(shadow, jemx, ee, resid, nsrc, xx, yy, solut, norm_d, sumfun, 0);
      fitcount = 1;

      if (rms < 0.0) return(-1);	/* bad data */

      *sh_rms2 = rmscalc(resid, soft_gr);

      mini_rms = mini_const = 0;
      valmin_const = fabs(solut[0]);
      valmin_rms = rms;
      
      /* ******************************************************** */
      /*                                                          */
      /*    Backproject residual shadowgram to form sky image     */
      /*                                                          */
      /* ******************************************************** */
      
      backpro_d(resid, skyim_r);
      
      for (i=0; i<skydim; i++) fitsimage[ee][1][i] = skyim_r[i]; 

      clean(skyim_r, skyim);		/* remove large scale structures in skyimage */
         
      for (i=0; i<skydim; i++) fitsimage[ee][2][i] = skyim[i]; 

	/* after first fit output background cleaned map */ 
igen:
      /* ******************************************************** */
      /* ******************************************************** */
      /* *                                                      * */
      /* *      Iterative source removal loop starts here       * */
      /* *                                                      * */
      /* ******************************************************** */
      /* ******************************************************** */

      nsrc++;
      tours++;
      if (nsrc >= 19) {
/*         printf(" ############ stop due to nsrc = 19\n"); 
*/
   	 putchar((char)('0'+1));
         goto nextee;	/* next energy range */
      } 
               
      /* ******************************************************** */
      /*                                                          */
      /*            Find dominant source in sky image             */
      /*                                                          */
      /* ******************************************************** */

      if (blindnum > 0) {
         for (i=0; i<blindnum; i++) {
            j= blindspots[i];
            skyim[j] = 0.0;
            skyim[j+256] = 0.0;
            skyim[j-256] = 0.0;
            skyim[j+257] = 0.0;
            skyim[j-257] = 0.0;
            skyim[j+255] = 0.0;
            skyim[j-255] = 0.0;
            skyim[j + 1] = 0.0;
            skyim[j - 1] = 0.0;
            skyim[j+512] = 0.0;
            skyim[j-512] = 0.0;
            skyim[j+513] = 0.0;
            skyim[j-513] = 0.0;
            skyim[j+514] = 0.0;
            skyim[j-514] = 0.0;
            skyim[j + 2] = 0.0;
            skyim[j - 2] = 0.0;
            skyim[j+511] = 0.0;
            skyim[j-511] = 0.0;
            skyim[j+510] = 0.0;
            skyim[j-510] = 0.0;
            skyim[j+258] = 0.0;
            skyim[j-258] = 0.0;
            skyim[j+254] = 0.0;
            skyim[j-254] = 0.0;
         }
      }

	/* findpeak finds the highest peak in the sky image */

      sxy = findpeak(skyim, access, smx, smy, amx, edge);
      
      if (sxy < 0) {
         putchar((char)('0'+7)); 
/*       printf(" ############ stop for bad resiual map \n");  */
   	 if (iter > 2) {
            ngrp--; 
            nsrc--; 	/* disregard last found source */
         }
         goto nextee;
      }
      
      if ((nsrc-backmodels) >= iter) {
  
         putchar((char)('0'+2)); 

/*       printf(" ############ stop for iter\n"); 
*/
   	 if (iter > 2) {
            ngrp--; 
            nsrc--; 	/* disregard last found source */
         }
         goto nextee;
      }
      if (tours >= 20) {
         putchar((char)('0'+3)); 
      
/*       printf(" ############ stop for more than 20 fit cycles\n"); 
*/  
   	 ngrp--; 
         nsrc--; 	/* disregard last found source */
         goto nextee;
      }
      if (amx[0] < (3.0 * amx[max_candidates])) {

/*       printf(" ############ stop for peak significance\n"); 
*/
   	 ngrp--; 
         nsrc--; 	/* disregard last found source */
         putchar((char)('0'+4));
         goto nextee;	/* check significance of new peak */
      }
      
      xnow = amx[1] * depth_corr;
      ynow = amx[2] * depth_corr;

      ngrp = nsrc - 1;
      xx[ngrp] = xnow;
      yy[ngrp] = ynow;
      sxy_list[ngrp] = sxy;
      updates[ngrp] = 0;
      
/*      printf("Dominant source position:  sxy: %5d,  XY: %+4d %+4d, delta xy: %+4d, %+2d, ngrp: %2d, fits %d\n",
         sxy, (int)(sxy/255 - 127), (int)((sxy%255) - 127), smx[1], smy[1], ngrp, fitcount[ee]);
*/
      just_updated = 0;

      /* ******************************************************** */
      /*                                                          */
      /*   Fit shadowgram with background and source components   */
      /* Generate residual shadowgram for next findpeak operation */
      /*                                                          */
      /* ******************************************************** */
newfit:      
      rms = imagefit(shadow, jemx, ee, resid, nsrc, xx, yy, solut, norm_d, sumfun, 0);
      fitcount++;
/*         printf("\n peak peak power: %8.3lf, counts: %10.1f\n", solut[ngrp], shadow_counts); 
*/      
      
      if (rms < 0.0) {
/*         printf(" ############ stop for negative rms"); 
*/
         putchar((char)('0'+5));
         goto nextee;
      }
      
      if (just_updated == 1) goto no_update_check;
      
      	/* check for previously found peak */ 
      if (ngrp-backmodels > 0) {
         for (m=backmodels; m<ngrp; m++) {	
            if (((xnow-xx[m])*(xnow-xx[m])+(ynow-yy[m])*(ynow-yy[m])) < 6.0) {
               if (solut[ngrp] >= 0.0) {
                  if (solut[m] >= 0.0) {
                     xx[m] = (xx[m] * solut[m] + xnow * solut[ngrp]) / (solut[m] + solut[ngrp]);
                     yy[m] = (yy[m] * solut[m] + ynow * solut[ngrp]) / (solut[m] + solut[ngrp]);
                  } else {
                     xx[m] = xnow + (xnow - xx[m]) * solut[m] / (fabs(solut[m]) + solut[ngrp]);
                     yy[m] = ynow + (ynow - yy[m]) * solut[m] / (fabs(solut[m]) + solut[ngrp]);
                  }
               } else {
                  if (solut[m] >= 0.0) {
                     xx[m] = xx[m] + (xnow - xx[m]) * solut[ngrp] / (solut[m] - fabs(solut[ngrp]));
                     yy[m] = yy[m] + (ynow - yy[m]) * solut[ngrp] / (solut[m] - fabs(solut[ngrp]));
                  } else {
                     xx[m] = (xx[m] * solut[m] + xnow * solut[ngrp]) / (solut[m] + solut[ngrp]);
                     yy[m] = (yy[m] * solut[m] + ynow * solut[ngrp]) / (solut[m] + solut[ngrp]);
                  }
               }
               sxy = sxy_list[m];
               ngrp--;
               nsrc--;
               updates[m]++;
               just_updated = 1;
               if (updates[m] >= 5) {
                  blindspots[blindnum] = sxy;
                  blindnum++;
               }
/*               printf("%d update for source %1d position:  sxy: %5d,  XY: %6.1f %6.1f, ngrp: %2d, blind: %d\n",
                  updates[m], m, sxy, xx[m], yy[m], ngrp, blindnum);
*/
               goto newfit;
            } 
         }
         for (m=backmodels; m<=ngrp; m++) updates[m] = 0;

         if (ngrp > backmodels) {
            if (rms < valmin_rms) valmin_rms = rms;
            else mini_rms++;
            if (fabs(solut[0]) < valmin_const) valmin_const = fabs(solut[0]);
            else mini_const++;
         } else {
            if (rms < valmin_rms) valmin_rms = rms;
            if (fabs(solut[0]) < valmin_const) valmin_const = fabs(solut[0]);
         }

no_update_check:
/*         printf("Update goodbet (if not past minimum: rms# %1d const: %1d): \n", mini_rms, mini_const);
*/
         if ((mini_rms == 0) || (mini_const == 0)) {
            for (m=0; m<=ngrp; m++) goodbet_g[m] = solut[m];
            *goodnsrc = nsrc;
            *goodrms = rms;
            *goodfits = fitcount;
         } else {
/*            printf(" ############ stop beyond best fit\n"); 
*/
   	    putchar((char)('0'+6));
            goto nextee;
         }  
         if (solut[ngrp] < limit_fraction * shadow_counts) {
/*     	    printf("\n ############ stop for peak power:  solut[ngrp]: %8.3lf, counts: %10.1f\n",
               solut[ngrp], shadow_counts); 
*/
   	    putchar((char)('0'+8));
            goto nextee;
         }
      
      }

      /* ******************************************************** */
      /*                                                          */
      /*    Backproject residual shadowgram to form sky image     */
      /*                                                          */
      /* ******************************************************** */

just_eliminate:      
      backpro_d(resid, skyim_r);

      clean(skyim_r, skyim);		/* remove large scale structures in skyimage */
         
      /* ******************************************************** */
      /*                                                          */
      /*  Repeat source finding and elimination from shadowgram   */
      /*                                                          */
      /* ******************************************************** */
      
      goto igen;

nextee: return(nsrc);
}
