#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <string.h>

#include "midi_b.h"

double longspx[6][2], longspy[6][2], longspw, longspw2, longsph, offs_lsp; 
double shortspx[6][2], shortspy[6][2], shortspw, shortspw2, shortsph, offs_ssp; 
double ringri[3], ringro[3], ringh[3], rx[4]; 
double ringri2[3], ringro2[3]; 
double bolt1x[6], bolt1y[6], bolt1r, bolt1s, offs_b1, bolth; 
double bolt2x[6], bolt2y[6], bolt2r, bolt2s, offs_b2; 
double bolt3x[6], bolt3y[6], bolt3r, bolt3s, offs_b3; 
double joint1x[6], joint1y[6], joint1r, joint1h, joint1r2, offs_j1; 
double joint2x[6], joint2y[6], joint2r, joint2h, joint2r2, offs_j2;
double twist, r_twist, pi6;
unsigned char struk[1100][1100];

void strukdef(void);
void s_turn(double xin, double yin, double angle, double *xout, double *yout, int prin);
void struktu(void);
void bar(double endx[2], double endy[2], double w, int mark);
void ring(double ri, double ro, int mark);
void bolt(double xc, double yc, double r, int mark, int prin);
int circle(double xc, double yc, double r, double x, double y, double *rv);
void strucprint(double offx, double offy);
void strucpr2(int step, int offx);
int struchk(double xm, double ym, double x_det, double y_det, int last);
void masktest(void);
   
/* *************************************************************** */

void struktu(void)
{
   int i, j, k, mark;
   double endx[12][2], endy[12][2], w[12]; 
   double rir[3], ror[3], onax; 
   double xc[30], yc[30], r[30]; 
   
   mark = 1;				/* mark for wide shadow border */
   ring(ringri2[0], ringro2[0], mark);	/* ring 1 shadow region (check at 21 mm) */
   mark = 0;				/* mark for narrow shadow border */
   ring(ringri2[1], ringro2[1], mark);	/* ring 2 shadow region (check at 10 mm) */
   ring(ringri2[2], ringro2[2], mark);	/* ring 3 shadow region (check at 10 mm) */

   for (i=0; i<6; i++) {
      mark = 1;		/* mark for wide shadow border */
      bar(longspx[i], longspy[i], longspw2, mark);
      bolt(bolt1x[i], bolt1y[i], bolt1s, mark,1);
      bolt(bolt2x[i], bolt2y[i], bolt2s, mark,1);
      mark = 0;		/* mark for narrow shadow border */
      bolt(bolt3x[i], bolt3y[i], bolt3s, mark,1);
      bar(shortspx[i], shortspy[i], shortspw2, mark);
      bolt(joint1x[i], joint1y[i], joint1r2, mark,1);
      bolt(joint2x[i], joint2y[i], joint2r2, mark,1);
   }

   mark = 64;		/* mark for exclusion */
   ring(ringri[0], ringro[0], mark);	/* mark ring 1 in struk array (excluded) */
   mark = 32;		/* mark for exclusion at 10 mm */
   ring(ringri[1], ringro[1], mark);	/* mark ring 2 (exclude at 10 mm) */
   ring(ringri[2], ringro[2], mark);	/* mark ring 3 (exclude at 10 mm) */
   
   for (i=0; i<6; i++) {
      mark = 64;		/* mark for exclusion */
      bar(longspx[i], longspy[i], longspw, mark);
      bolt(bolt1x[i], bolt1y[i], bolt1r, mark,1);
      bolt(bolt2x[i], bolt2y[i], bolt2r, mark,1);

      mark = 32;		/* mark solid regions for exclusion at 10 mm) */
      bar(shortspx[i], shortspy[i], shortspw, mark);
      bolt(joint1x[i], joint1y[i], joint1r, mark, 1);
      bolt(joint2x[i], joint2y[i], joint2r, mark, 1);
      bolt(bolt3x[i], bolt3y[i], bolt3r, mark,1);
   }

   k = 0;
   for (i=-250; i<250; i++) {   
      for (j=-250; j<250; j++) {   
         if ((i*i+j*j) > 62500) continue;
         if (struk[550+i][550+j] >= 32) k++;
      }
   }
   
   onax = (double)(k) / (3.14159 * 250.0 * 250.0);
   
/*   printf("\nOn-axis blocking by support structure: %6.3lf\n", onax);

      strucpr2(1, 0);
      strucpr2(1, 250);
      strucpr2(2, 0);
      strucpr2(2, 150);
      strucpr2(2, 300);
      strucpr2(2, 400);
      strucpr2(4, 0);
      strucpr2(8, 0);
*/	   																																		
   return;
}
	
/* *************************************************************** */

void bar(double endx[2], double endy[2], double w, int mark)
{
   int i, j;
   double length, xv, yv, dx, dy, r, rin, rout;
   
   length = sqrt((endx[0]-endx[1])*(endx[0]-endx[1]) +
      		 (endy[0]-endy[1])*(endy[0]-endy[1]));
   rin = sqrt(endx[0]*endx[0] + endy[0]*endy[0]);
   rout = sqrt(endx[1]*endx[1] + endy[1]*endy[1]);
/*   printf("Bar length: %7.2lf, rin: %7.2lf, rout: %7.2lf, endxy: %7.2lf/%7.2lf\n",
      		length, rin, rout, endx[0], endy[0]);
*/
   r = w / 2.0;
   
   j = length / 0.5 + 1;
   dx = (endx[0]-endx[1]) / j;
   dy = (endy[0]-endy[1]) / j;
   for (i=0; i<=j; i++) {
      xv = endx[1] + i * dx;
      yv = endy[1] + i * dy;
      bolt(xv, yv, r, mark, 0);
   }
   
/*   strucprint((endx[0]+endx[1])/2.0, (endy[0]+endy[1])/2.0);
*/
   return;
}
	
/* *************************************************************** */

void ring(double ri, double ro, int mark)
{
   int ix, iy, xlow, xhigh, ylow, yhigh, n;
   double xv, yv, dn = 0.0, rc, rv;
   
   xlow = 550 - ro * 2.0;
   xhigh = 550 + ro * 2.0;
   ylow = 550 - ro * 2.0;
   yhigh = 550 + ro * 2.0;
   rc = (ri + ro) / 2.0;
   
   if (xlow < 0) xlow = 0;
   if (xhigh > 1099) xhigh = 1099;
   if (ylow < 0) ylow = 0;
   if (yhigh > 1099) yhigh = 1099;
   n = 0;
   
   for (ix=xlow; ix<=xhigh; ix++) {
      for (iy=ylow; iy<=yhigh; iy++) {
         xv = (double)(ix - 550) / 2.0;
         yv = (double)(iy - 550) / 2.0;
         if (circle(dn, dn, ro, xv, yv, &rv) == 1) {
   	    if (rv > ri) {
               if (mark <= 1) struk[ix][iy] = strukupdate(xv, yv, rv) + mark; 
               if (mark > 1) struk[ix][iy] = mark;
               n++;
            }
         }
      }
   }
/*   strucprint(rc, 0.0);
   strucprint(0.0, rc);
   strucprint(-rc, 0.0);
   strucprint(0.0, -rc);
*/
/*   printf("Ring: ri/ro: %6.2lf/%6.2lf, number of affected pixels: %6d,  mark: %1d\n", ri, ro, n, mark);
*/
   return;
}
	
/* *************************************************************** */

void bolt(double xc, double yc, double r, int mark, int prin)
{
   int ix, iy, ixlow, ixhigh, iylow, iyhigh, mrk;
   double xv, yv, c_angle, angle, rv, rb;
   
   ixlow = (xc - r) * 2.0 + 550.0;
   ixhigh = (xc + r) * 2.0 + 550.0;
   iylow = (yc - r) * 2.0 + 550.0;
   iyhigh = (yc + r) * 2.0 + 550.0;

   if (ixlow < 0) ixlow = 0;
   if (ixhigh > 1099) ixhigh = 1099;
   if (iylow < 0) iylow = 0;
   if (iyhigh > 1099) iyhigh = 1099;
   for (ix=ixlow; ix<=ixhigh; ix++) {
      for (iy=iylow; iy<=iyhigh; iy++) {
         xv = (double)(ix - 550) / 2.0;
         yv = (double)(iy - 550) / 2.0;
         if (circle(xc, yc, r, xv, yv, &rb) == 1) {
   	    if (mark > 1) struk[ix][iy] = mark;
            if (mark <= 1) {
               rv = sqrt(xv*xv + yv*yv);
               struk[ix][iy] = strukupdate(xv, yv, rv) + mark; 
            }
         }
      }
   }
/*   if (prin == 1) strucprint(xc, yc);
*/
   return;
}

/* *************************************************************** */

int circle(double xc, double yc, double r, double x, double y, double *rv)
{
   int inside;
   static int m=0;
   
   inside = 0;
   
   if ((*rv = sqrt((x-xc)*(x-xc) + (y-yc)*(y-yc))) <= r) inside = 1;
   m++;
   
/*   if (m < 1000) {
      printf("Circle: m: %3d, xc/yc: %5.1lf/%5.1lf, r: %5.1lf, x/y: %6.1lf/%6.1lf, inside: %1d\n",
         m, xc, yc, r, x, y, inside);
   }
*/   
   return(inside);
}

/* *************************************************************** */

void strukdef(void)
{
   int i, j, s_offi, s_offj, prin;
   double ta, len;
   
/*   printf("Entering strukdef\n");
*/
   twist = -4.286;		/* twist angle for entire support structure */
   r_twist = twist * gtorad;
   pi6 = pi / 6.0;
   for (i=0; i<1100; i++) for (j=0; j<1100; j++) struk[i][j] = 0;

   ringri[0] = 45.0;		/* inner radius of ring 1 */
   ringro[0] = 48.0;		/* outer radius of ring 1 */
   ringri2[0] = 45.0 - 3.5;	/* inner radius of shadow region */
   ringro2[0] = 48.0 + 3.5;	/* outer radius of shadow region */
   ringri[1] = 114.9;		/* inner radius of ring 2 */
   ringro[1] = 117.1;		/* outer radius of ring 2 */
   ringri2[1] = 114.9 - 1.7;	/* inner radius of shadow region */
   ringro2[1] = 117.1 + 1.7;	/* outer radius of shadow region */
   ringri[2] = 210.0;		/* inner radius of ring 3 */
   ringro[2] = 212.2;		/* outer radius of ring 3 */
   ringri2[2] = 210.0 - 1.7;	/* inner radius of shadow region */
   ringro2[2] = 212.2 + 1.7;	/* outer radius of shadow region */
   rx[0] = (ringri[0] + ringro[0]) / 2.0;
   rx[1] = (ringri[1] + ringro[1]) / 2.0;
   rx[2] = (ringri[2] + ringro[2]) / 2.0;
   rx[3] = 267.5;
   
   /* NOTE: the initial coordinates are not yet twisted by -4.286 degrees ! */
   longspx[0][0] = 46.5;	/* inner end of long spoke */
   longspx[0][1] = 268.0;	/* outer end of long spoke */
   longspy[0][0] = 0.0;		/* inner end of long spoke */
   longspy[0][1] = 0.0;		/* outer end of long spoke */
   longspw = 2.2;		/* width of long spoke */
   longspw2 = 2.2 + 7.0;	/* width of long spoke shadow */
   longsph = 21.25;		/* height of long spoke */
   offs_lsp = 30.0 + twist;	/* angular offset of first long spoke */

   shortspx[0][0] = 116.0;	/* inner end of short spoke */
   shortspx[0][1] = 211.0;	/* outer end of short spoke */
   shortspy[0][0] = 0.0;	/* inner end of short spoke */
   shortspy[0][1] = 0.0;	/* outer end of short spoke */
   shortspw = 2.2;		/* width of short spoke */
   shortspw2 = 2.2 + 3.4;	/* width of shadow region */
   shortsph = 10.25;		/* height of short spoke */
   offs_ssp = 0.0 + twist;	/* angular offset of first short spoke */

   bolt1x[0] = 49.5;		/* x coordinate of bolt1 */
   bolt1y[0] = 0.0;		/* y coordinate of bolt1 */
   bolt1r = 4.5;		/* radius of bolt1 */
   bolt1s = 4.5 + 3.5;		/* radius of bolt1 shadow */
   offs_b1 = 30.0 + twist;	/* angular offset of first bolt in 1. bolt set */

   bolt2x[0] = 116.0;	/* x coordinate of bolt2 */
   bolt2y[0] = 0.0;		/* y coordinate of bolt2 */
   bolt2r = 4.5;		/* radius of bolt */
   bolt2s = 4.5 + 3.5;		/* radius of bolt2 shadow */
   offs_b2 = 30.0 + twist;	/* angular offset of first bolt in 2. bolt set */
   
   bolt3x[0] = 211.4;	/* x coordinate of bolt3 */
   bolt3y[0] = 0.0;		/* y coordinate of bolt3 */
   bolt3r = 3.5;		/* radius of bolt */
   bolt3s = 3.5 + 3.5;		/* radius of bolt3 shadow */
   offs_b3 = 0.0 + twist;	/* angular offset of first bolt in 3. bolt set */
   
   joint1x[0] = 211.4;	/* x coordinate of joint1 */
   joint1y[0] = 0.0;		/* y coordinate of joint1 */
   joint1r = 3.34;		/* radius of joint1 */
   joint1r2 = 3.34 + 1.7;	/* outer radius of shadow region */
   offs_j1 = 30.0 + twist;	/* angular offset of first joint in 1. joint set */
   
   joint2x[0] = 117.1;	/* x coordinate of joint2 */
   joint2y[0] = 0.0;		/* y coordinate of joint2 */
   joint2r = 2.5;		/* radius of joint2 */
   joint2r2 = 2.5 + 1.7;	/* radius of shadow region */
   offs_j2 = 0.0 + twist;	/* angular offset of first joint in 2. joint set */
   
   for (j=1; j<7; j++) {  
      i = j % 6;
      ta = 60.0 * i;
/*      if ((i%3) == 0) prin = 1; else prin = 0;
*/
      s_turn(longspx[0][0], longspy[0][0], ta+offs_lsp, &longspx[i][0], &longspy[i][0], prin);  
      s_turn(longspx[0][1], longspy[0][1], ta+offs_lsp, &longspx[i][1], &longspy[i][1], prin);  
/*      len = sqrt((longspx[i][0]-longspx[i][1]) * (longspx[i][0]-longspx[i][1]) +
         	 (longspy[i][0]-longspy[i][1]) * (longspy[i][0]-longspy[i][1]));
      printf("long %1d: x1/y1: %7.2lf/%7.2lf,  x2/y2: %7.2lf/%7.2lf,  length: %7.2lf\n", 
         i, longspx[i][0], longspy[i][0], longspx[i][1], longspy[i][1], len);
*/
      s_turn(shortspx[0][0], shortspy[0][0], ta+offs_ssp, &shortspx[i][0], &shortspy[i][0], prin);  
      s_turn(shortspx[0][1], shortspy[0][1], ta+offs_ssp, &shortspx[i][1], &shortspy[i][1], prin);  
/*      len = sqrt((shortspx[i][0]-shortspx[i][1]) * (shortspx[i][0]-shortspx[i][1]) +
         	 (shortspy[i][0]-shortspy[i][1]) * (shortspy[i][0]-shortspy[i][1]));
      printf("short %1d: x1/y1: %7.2lf/%7.2lf,  x2/y2: %7.2lf/%7.2lf,  length: %7.2lf\n", 
         i, shortspx[i][0], shortspy[i][0], shortspx[i][1], shortspy[i][1], len);
*/
      prin = 0;
      s_turn(bolt1x[0], bolt1y[0], ta+offs_b1, &bolt1x[i], &bolt1y[i], prin);  
      s_turn(bolt2x[0], bolt2y[0], ta+offs_b2, &bolt2x[i], &bolt2y[i], prin);  
      s_turn(bolt3x[0], bolt3y[0], ta+offs_b3, &bolt3x[i], &bolt3y[i], prin);  
      s_turn(joint1x[0], joint1y[0], ta+offs_j1, &joint1x[i], &joint1y[i], prin);  
      s_turn(joint2x[0], joint2y[0], ta+offs_j2, &joint2x[i], &joint2y[i], prin);  
   }
   
   struktu();	/* build structure image array */
   

   return;
}

/* *************************************************************** */

void s_turn(double xin, double yin, double angle, double *xout, double *yout, int prin)
{
   double sin_a, cos_a; 

   sin_a = sin(angle*gtorad);
   cos_a = cos(angle*gtorad);
   
   *xout = cos_a * xin - sin_a * yin; 
   *yout = cos_a * yin + sin_a * xin; 
   
   if (prin == 1) {
      printf("angle: %6.2lf, sin %6.4lf, cos %6.4lf, x/y in: %6.2lf/%6.2lf, x/y out: %6.2lf/%6.2lf\n", 
         	angle, sin_a, cos_a, xin, yin, *xout, *yout); 
   }
   return;
}

/* *************************************************************** */

void strucprint(double offx, double offy)
{
   int i, j, s_offi, s_offj;
   char princhar;

   s_offi = 550 + (int)(offy * 2.0 + 0.5);
   s_offj = 550 + (int)(offx * 2.0 + 0.5);
   if (s_offi < 21) s_offi = 21;
   if (s_offi > 1078) s_offi = 1078;
   if (s_offj < 64) s_offj = 64;
   if (s_offj > 1035) s_offj = 1035;

   printf("offx: %7.1lf, offy: %7.1lf,  off_ix: %5d, off_iy: %5d\n", offx, offy, s_offi, s_offj);
   for (i=21; i>=-21; i--) {
      printf("\n");
      for (j=-64; j<=64; j++) {
         if (struk[s_offj+j][s_offi+i] < 10) princhar = '0' + struk[s_offj+j][s_offi+i];
         else princhar = 'A' + struk[s_offj+j][s_offi+i] - 10;
         printf("%1c", princhar);
/*         if (struk[s_offj+j][s_offi+i] > 0) printf("%1x", struk[s_offj+j][s_offi+i]);
         else printf(".");
*/
      }
   }
   printf("\n\n");
}   

/* *************************************************************** */

void strucpr2(int step, int off)
{
   int i, j, s_offi, s_offj;
   char princhar;

   s_offi = 550 + off;
   s_offj = 550;
   if (s_offi < 64*step) s_offi = 64 * step;
   if (s_offi > (1099 - 64*step)) s_offi = 1099 - 64*step;
   if (s_offj < 64*step) s_offj = 64 * step;
   if (s_offj > (1099 - 64*step)) s_offj = 1099 - 64*step;

   printf("offx: 0.0, offy: 0.0,  off_ix: 550, off_iy: 550,  step: %2d \n", step);
   for (i=64*step; i>=-64*step; i-=step) {
      printf("\n");
      for (j=-64*step; j<=64*step; j+=step) {
         if (struk[s_offj+j][s_offi+i] < 10) princhar = '0' + struk[s_offj+j][s_offi+i];
         else if (struk[s_offj+j][s_offi+i] < 64) princhar = 'A' + (struk[s_offj+j][s_offi+i] & 0x1F) - 10;
              else princhar = '';
         if (struk[s_offj+j][s_offi+i] == 0) {            
            princhar = ' ';
            if (((i/step) % 16) == 0) princhar = '-';
            if (((j/step) % 16) == 0) princhar = '|';
         }
         printf("%1c", princhar);
/*         if (struk[s_offj+j][s_offi+i] > 0) printf("%1x", struk[s_offj+j][s_offi+i]);
         else printf(".");
*/
      }
   }
   printf("\n\n");
}   

/* *************************************************************** */

int strukupdate(double xv, double yv, double rv) 
{
   int m6, m12;
   
   if (rv < rx[0]) return(0);
   
   m12 = (atan2(yv, xv) + twopi - r_twist) / pi6;
   m12 %= 2;
   m6 = (atan2(yv, xv) + twopi - r_twist - pi6) / (pi6 * 2.0);
   m6 %= 2;
/*   if (rv < rx[1]) return(8 + m6);
   if (rv < rx[2]) return(16 + m12);
   if (rv < rx[3]) return(32 + m6);
*/
   if (rv < rx[1]) return(2+2*m6);
   if (rv < rx[2]) return(8+8*m12);
   if (rv < rx[3]) return(2+2*m6);
   return(0);    
}

/* *************************************************************** */

/*int struchk(double xm, double ym, double dxl, double dyl, double dxh, double dyh, int last)*/
int struchk(double xm, double ym, double x_det, double y_det, int last)
/*	The "struk" array contains information about the mask support structure and its
        shadowing of the mask. The array covers a 550x550 mm2 area with pixels of 0.5x0.5 mm2.
        The access tests should be done in the following order:
        1) calculate intersection of ray with the mask plane. 
           Find corresponding pixel in the "struk" array.

        2) If the pixel value is 0 access is allowed along this path     (in open areas between structural members)
        3) If the pixel value is >= 32 access is forbidden along this path (hit solid structure or mask elim. hole)
        4) If the pixel value is >0 and <32:                (in potential shadow area close to a structural member)
           4a) If the pixel value is odd, 		             (close to a structural member of 21 mm height)
              Calculate intersection points of this photon with two leves at +/- 21.25 mm
           4b) If the pixel value is even, 		             (close to a structural member of 10 mm height)
              Calculate intersection points of this photon with two leves at +/- 10.25 mm
        5) If the pixel value in any of the two new points is 64 access is forbidden.         (hit solid structure)
        6) If the pixel value in any of the two new points is 32 
           			and the first pixel value was even, access is forbidden.      (hit solid structure)
        7) If the pixel value in both of the two new points is >= 0 
                                and the two values are different, access is forbidden.   (traverse solid structure)
        8) Otherwise access is allowed.

        The following symbols are used in the pixels of struk:
           0:	free access, no further tests are required.
           32:  access blocked by solid material in mask support structure of 10.25 mm height.
           64:  access blocked by solid material in mask support structure of 21.25 mm height.
           1, 3, 5, 7, 9: 17: Photon close to high structure. 
              		  Test required at plus/minus 21.25 mm from the mask plane.
           2, 4, 6, 8, 16: Photon close to low structure. 
              		  Test required at plus/minus 10.25 mm from the mask plane.
*/
{
   int i, j, ix, iy, ixb, iyb, ixu, iyu, sbval, suval;
   double xb, yb, xu, yu, dxl, dyl, dxh, dyh, xdif, ydif;
   static int p_count[10] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, first = 1;
   
   if (first) {
         /* check that the mask_height has been defined (external variable) */
      if ((mask_height < 3390.0) || (mask_height > 3403.0)) {
         printf("In m_strukdef: mask_height value outside acceptable range: %6.2lf\n", mask_height);
         exit(0);
      }
      strukdef();	/* initialize structure definitions */
      first = 0;
   }
   
   xb = xu = xm * 2.0 + 550.5; 
   yb = yu = ym * 2.0 + 550.5;
   
   xdif = xm - x_det;
   ydif = ym - y_det;
   
   dxl = xdif * 2.0 * shortsph / mask_height; 
   dyl = ydif * 2.0 * shortsph / mask_height; 
   dxh = xdif * 2.0 * longsph / mask_height; 
   dyh = ydif * 2.0 * longsph / mask_height; 
   
   ix = xb;
   iy = yb;
   
   if ((ix < 0) || (ix > 1099) || (iy < 0) || (iy > 1099)) {
      printf("ix/iy outside limits: %4d/%4d, x/y_mask: %6.2f/%6.2f, x/y_det: %6.2f/%6.2f\n",
               ix, iy, xm, ym, x_det, y_det);
      ix = iy = 0;   
   }
   i = 0;
   if (struk[ix][iy] == 0) {
      if (last) p_count[7]++;
      goto retu;	/* in open areas between structural members */
   }
   i = 1;
   if (struk[ix][iy] >= 32) {
      if (last) p_count[0]++;
      goto retu;	/* hit solid structure */
   }
   if ((struk[ix][iy]%2) == 0) {	/* struk value is even - close to a structural member of 10 mm height */
      xb -= dxl;
      yb -= dyl;
      ixb = xb;
      iyb = yb;

      if ((ixb < 0) || (ixb > 1099) || (iyb < 0) || (iyb > 1099)) {
         printf("ixb/iyb outside limits: %4d/%4d, x/y_mask: %6.2f/%6.2f, x/y_det: %6.2f/%6.2f\n",
               ixb, iyb, xm, ym, x_det, y_det);
         ixb = iyb = 0;   
      }
      sbval = struk[ixb][iyb];
      if (sbval >= 32) {
         if (last) p_count[1]++;
         goto retu;	/* hit solid structure */
      }
   
      xu += dxl;
      yu += dyl;
      ixu = xu;
      iyu = yu;
      if ((ixu < 0) || (ixu > 1099) || (iyu < 0) || (iyu > 1099)) {
         printf("ixu/iyu outside limits: %4d/%4d, x/y_mask: %6.2f/%6.2f, x/y_det: %6.2f/%6.2f\n",
               ixu, iyu, xm, ym, x_det, y_det);
         ixu = iyu = 0;   
      }
      suval = struk[ixu][iyu];

      if (suval >= 32) {
         if (last) p_count[2]++;
         goto retu;	/* hit solid structure */
      }
      if ((sbval > 0) && (suval > 0) && ((sbval&0xffe) != (suval&0xffe))) {
         if (last) p_count[3]++;
         goto retu;	/* traverse structure */
      }
   } else {				/* struk value is odd - close to a structural member of 21 mm height */

      xb = xu = xm * 2.0 + 550.5; 
      yb = yu = ym * 2.0 + 550.5;
   
      xb -= dxh;
      yb -= dyh;
      ixb = xb;
      iyb = yb;

      if ((ixb < 0) || (ixb > 1099) || (iyb < 0) || (iyb > 1099)) {
         printf("ixb/iyb2 outside limits: %4d/%4d, x/y_mask: %6.2f/%6.2f, x/y_det: %6.2f/%6.2f\n",
               ixb, iyb, xm, ym, x_det, y_det);
         ixb = iyb = 0;   
      }
      sbval = struk[ixb][iyb];
      if (sbval >= 64) {
         if (last) p_count[4]++;
         goto retu;	/* hit solid structure */
      }
   
      xu += dxh;
      yu += dyh;
      ixu = xu;
      iyu = yu;
      if ((ixu < 0) || (ixu > 1099) || (iyu < 0) || (iyu > 1099)) {
         printf("ixu/iyu2 outside limits: %4d/%4d, x/y_mask: %6.2f/%6.2f, x/y_det: %6.2f/%6.2f\n",
               ixu, iyu, xm, ym, x_det, y_det);
         ixu = iyu = 0;   
      }
      suval = struk[ixu][iyu];
      if (suval >= 64) {
         if (last) p_count[5]++;
         goto retu;	/* hit solid structure */
      }
      if ((sbval > 0) && (suval > 0) && ((sbval&0xffe) != (suval&0xffe))) {
         if (last) p_count[6]++;
         goto retu;	/* traverse structure */
      }
   }   
   if (last) p_count[8]++;
   i = 0;
retu:

   if (last < 2) if (i > 0) return(1); else return(0);
   printf("Strukdef statistics:\n");
   printf(" msk_level  low_under  low_over  low_trav  hgh_under  hgh_over  hgh_trav  init_OK  final_OK\n");
   for (j=0; j<9; j++) {
      printf("%8d  ", p_count[j]);
      p_count[j] = 0;
   }
   printf("\n");

   if (i > 0) return(1); else return(0);
}

