/*
Shadowgram backprojection. Niels Lund Dec. 2002.
Version 2.0 (minisky) Jan 2003 Using empirical empty field background maps and fixed sky-pixel source models.
Version 3.0 (midisky) Feb 2003 Using pixelfolding of shadowgrams and calculating source models according
                               to interpixel position determined by findpeak.
Uses pre-calculated shadowgrams "shadgram" in four energy bands. (prepared by "ipfread")  [0ymmdd_#shdJ.bin]
Uses a pre-calculated sky-backprojection file - with auxiliary data. (prepared by "skymapk")
Uses smoothed empty-field shadowgrams "soft_shdgr" in four energy bands. (prepared by "b_modelsmo")
Version 4.0 (midisky6-7) Uses pointing data to calculate sky cordinates.
                         (Programme "skyadjust.c" have been used to optimize conversion from 
                          instrument to sky-coordinates)
Version 5.0 (midisky8) Introduces lists of all source candidates seen any science window for a full file
   			and search for consistent localizations in this list (to catch low energy sources)
Version 6.0 (midiskyb, Aug. 2003) Take mask support structure and eliminated holes into account
midi_b3 Version 1.0 (Sep. 2003) Improve structure by splitting into separate files an eliminating dead code
midi_b4 Version 2.0 (Oct. 2003) Include generation of sky images in fits format
midi_b5 Version 1.0 (Nov. 2003) Include control data reading from files
midi_b5 Version 1.1 (March 2004) Export version with various minor updates & corrections
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/timeb.h>

#include "midi_b.h"

double pi, twopi, gtorad;

/* !!! The following are backprojection related data */

int d_allow_p[detdim];
short d_allow_l[detdim];
unsigned d_allow_st[detdim];
unsigned char d_allow[max_sky_bytes];
float soft_gr[detdim];
int det_e[skydim];
double soft_tot[30];
int FOV_radius = 127; 
int jmxsign, o_ctrl;

char access[skydim];
float rsignif[skydim];
float fitsimage[4][3][skydim], skyim_r[skydim], skyim[skydim];
 
float soft_shdgr[4][detdim];
float soft_gr1[4][detdim];
double soft_tot1[30];

short shadgram[4][detdim];
double shdgr_aux[30];
double shdgr_aux1[30];
short maskhole[22501];		/* 22501 is the "holy" prime number on which the JEM-X mask is based */ 

double parameters[20];	/* array with the specific  parameters for current skymap:
        parameters[0] = byt_index;	length of 'd_allow' list 

        parameters[7] = cond;		condition parameter for inclusion of collimator in backprojection 
	          		cond = 0: disregard collimator in backprojection 
                                cond = 1: include collimator
        parameters[8] = jemx; 		JEM-X unit number 
        parameters[9] = FOV_radius	Radius of FOV expressed in sky-pixels
        parameters[10] = indep;		interaction depth of photons in the detector 
           			Common-X is the direction from JEM-X2 towards JEM-X1 
        parameters[11] = alfa; 		angle between mask X-axis and Common-X 
        parameters[12] = beta;		angle between collimator X-axis and Common-X  		
        parameters[13] = gamma;		angle between microstrip X-axis and Common-X 

        parameters[14] = 0.0;		was 'Hsource' for ESTEC calibrations, now not used 
        parameters[15] = xshi;		X-shift of microstrip center relative to collimator center 
        parameters[16] = yshi;		Y-shift of microstrip center relative to collimator center 
        parameters[17] = maskoffs;	displacement of mask center spacecraft centerline, now not used
        parameters[18] = mask_height;	height of mask above detector window 
        parameters[19] = sky_version;	skymap version (must correspond to the "detmap" version)
*/

char notes[1000];	/* notes & comments for current skymap */

unsigned char pixel_list[500000];

double xshi, yshi, accumT[2], absolen, depth_corr, offQ, offR;
double scaleX, scaleY, offX, offY, x2x, y2y, phase2, j_RA, j_dec;

double abs_len[4] = {1.0, 4.0, 15.0, 55.0};	/* absorbtion length (mm) in the four energy bands */
						/* new values for low energy version */
double x_rms[4] = {1.11, 1.11, 0.53, 0.52}; 	/* position determination uncertainty */
double y_rms[4] = {1.03, 1.03, 0.49, 0.48}; 	/* new values for low energy version */

unsigned bb[32]; 

int scatt_delta[500], scatt_len, pixel_st[detdim];
int aux_orbit, aux_pid, aux_pidv, pha_sel[5];
int backmodels = 2, thisgroup = 0, signiflim = 100;

char sweethome[200];
char sweethome2[200];

float com_RA[sourcmax], com_dec[sourcmax], shadow[detdim], scatt_list[500];

double coll_r;		/* collimator radius */ 
double mask_r;		/* radius of coded mask pattern */
double mask_height;	/* mask-detector_window distance */
double indep;

FILE *cat_fil, *srclistf, *srclist2;

/* **************************************************************************************** */

int backsky_init(int jemx, int dead_ano, char sky_version);	/* initialize global parameters */
void peaktransfer(float x, float y, float imraw[skydim], float impeak[skydim]);
void fi_headr(char fitsfil[]);

/* **************************************************************************************** */
/* **************************************************************************************** */


main(int argc, char* argv[])
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*                                                                 */        
/*              midi_b, Niels Lund, September 2003                 */        
/*                                                                 */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

{
   char iname[100], filnavn[200], dfilnavn[100], dummy[540], skymapf[100], fitsnavn[128];
   char dfilnavn1[100], resultxt[150];
   unsigned char tmbyte[440];
   char sky_version = {'l'};
   
   float xx_g[2][5][sourcmax], yy_g[2][5][sourcmax], xx[sourcmax], yy[sourcmax];
   float goodbet_g[2][5][sourcmax], imgpeak[4][skydim], chkX, chkY;
   float norm_d[sourcmax], norm_d_g[2][5][sourcmax], backcnt[2][5][sourcmax];
   float final_s[2][sourcmax], norm_c[2][sourcmax], com_n[2][sourcmax];
   float com_avx[sourcmax], com_avy[sourcmax], com_sn[sourcmax], sumfun[sourcmax]; 
   float com_x[2][sourcmax], com_y[2][sourcmax], com_s[2][sourcmax];
   double p_RA[10000], p_dec[10000], p_posang[10000];   
   float shadow1[4][detdim], fk, sumsn, totsn, sn_best, sn_temp, temptot, tempgood;
   float tt, tt_old, tt_old2, sn_new, resid[detdim];
   float accum0x[accumax], accum1x[accumax], accum0y[accumax], accum1y[accumax], accums0[accumax];
   float accums1[accumax], accums[accumax];
   float accum2x[accumax], accum2y[accumax], accum3x[accumax], accum3y[accumax], acc2sn[accumax], acc3sn[accumax];
   float accu2RA[accumax], accu2dec[accumax], actot2[accumax], accums2[accumax];
   float actot0[accumax], actot1[accumax], accumRA[accumax], accumdec[accumax], accumx[accumax], accumy[accumax];
   float goodrms[2][5], valmin_rms, valmin_const, rms1, rms_old, cmx, cmy;
   float avr, cosdec, sumcnts;
   float j_RA, j_dec, sumX, sumY, sumgood, src_chk_px, src_chk_py;
   float s_dec, s_RA, phaX, phaY, avX0, avY0, edge;

   int i, j, ii, n, status, apid, jemx, jemx_s, sky_index, value, length, nsrc, oline=0, chk_no;
   int selectnum, det_index, ee, lmax = 18, pointval, bitpix, dim1, dim2, f_ctrl;
   int iterations, nsrc_g[2][5];
   int k, m, accu=0, fillen, fillen_old = 0, filnum=0, j_1[sourcmax];
   int starTm, starTh, stopTh, stopTm, dead_ano, newpoint, n_common, n_common2, src_coord=0, iee;
   int goodnsrc[2][5], goodfits[5];
   int totcnt[2][5], e1, e2, eee[2][sourcmax], ksrc, jb;
   int tours, q_plus, last_i; 
   double eval[5] = {2.4, 4.2, 8.4, 14.0, 35.0}; 
   int a_orbit[accumax], a_pid[accumax], a_pidv[accumax], accucode[accumax], jemxlocal;
   int mnum_g[2][5][sourcmax], accu2, accu3;
   int acc2pid[accumax], acc3pid[accumax], pidlast, orbit, pid, pidv, e_indic, acc3ee[accumax], ee0, ee1, ee2, ee3; 
   int acc3tot[accumax];

   short used[2][5][sourcmax], used1[20000], used2[20000], accumn0[accumax], accumn1[accumax], acc_i[accumax];
   short acc2ee[accumax], acc2jx[accumax];
   short shadow_r[detdim];
      
   double rms, rmsold, fitpwr, solut[sourcmax], limit_f, quality, qual_limit = 10.0;
   double solut_e[4][sourcmax], finalvar[4], sh_rms[5][4], sh_rms2;
   double startT, stopT, xrms, yrms, solutold; 
   double strength_g[2][sourcmax], sumtot, cdelt, rad_ang;

   double m_lst_ra[20000], m_lst_dec[20000], m_lst_x[20000], m_lst_y[20000], m_lst_s[20000], m_lst_T[20000];
   double m_lst_s0[20000], m_lst_s1[20000], m_lst_T0[20000], m_lst_T1[20000], m_lst_cnt[20000], m_lst_area[20000];
   double m_lst_q[20000];
   int m_lst_pid[20000], m_lst_e[20000], m_lst_jx[20000], m_lst_ee[20000], m_num=0;
   int s_reject[4] = {0, 0, 0, 0};
   int s_accept[4] = {0, 0, 0, 0};

   FILE *in, *gnu_fil, *fillist, *chkfil;
   
   Fun fun_handle;
   
   if ( argc > 1) strcpy(iname, argv[1]) ;	/* list of shadowgram files, 4 shadowgrams for one JEM-X */
   else {
      printf("Remember file_list! (Other control parameters must be in 'midi_b5_control.dat')\n");
      return;
   }

   cat_fil = fopen("accum.cat", "wt");

   sprintf ( sweethome, "/nfs/r6/jemx/nlwork/midi_distribution_2/work");
   sprintf ( sweethome2, "/nfs/r6/jemx/nlwork/midi_distribution_2/work/reference");
/*   sprintf(sweethome, "/home/nl/jemx/ipfread/jemx1-2");
   sprintf(sweethome2, "/home/nl/jemx/ipfread/jemx1-2/reference");
*/
   fprintf(cat_fil, "0  V10.4 20030603 corrected error in calculation of illuminated areas in 'imagefit'\n");
   fprintf(cat_fil, "0  V10.4 Modified calculation of SN (using only relevant fraction of the background)\n");
   fprintf(cat_fil, "0  V10.5 Include output of source model patterns  (midisky8) \n");
   fprintf(cat_fil, "0  V11 Include mask support structure and eliminated holes in modelling  (midiskyb) \n");
   fprintf(cat_fil, "0  V11 Uses sky version 'k' \n");
   fprintf(cat_fil, "0  midi_b V1. source code split into files, header file introduced. Uses sky version 'k'\n");
   fprintf(cat_fil, "0  midi_b4 V2. uses 'funtools' to output image files in fits format\n");
   fprintf(cat_fil, "0  midi_b5 V1. uses instrument data input from '--/reference/alignment.dat'\n");
   fprintf(cat_fil, "0  midi_b5 V1. and control data input from 'midi_b5_control.dat'\n");
   fclose(cat_fil);
   
   cat_fil = fopen("accum.cat", "at");
   
   sprintf(resultxt, "%s/reference/alignment.dat", sweethome);
   in = fopen(resultxt, "rt");
   if (in == NULL) {
      printf("No alignment file found! (%s)\n", resultxt);
      return;
   } 
      
   printf("0 Alignment file: %s \n", resultxt);
   for (i=0; i<200; i++) {
      fgets(resultxt, 150, in);
      if (feof(in)) goto q_end;
      fprintf(cat_fil, "0 Alignm. %s", resultxt);
   }
q_end:
   fclose(in);
   
   sprintf(resultxt, "midi_b5_control.dat");
   in = fopen(resultxt, "rt");
   if (in == NULL) sprintf(resultxt, "%s/reference/midi_b5_control.dat", sweethome);
   in = fopen(resultxt, "rt");
   if (in == NULL) {
      printf("No control data file found! (%s)\n", resultxt);
      return;
   } 
      
   printf("0 Control data file: %s \n", resultxt);
   for (i=0; i<200; i++) {
      fgets(resultxt, 150, in);
      if (feof(in)) goto x_end;
      fprintf(cat_fil, "0 Control %s", resultxt);
      printf("%s", resultxt);
      sscanf(resultxt, "%d", &n);
      if (n == 1) sscanf(resultxt, "%d %d", &n, &dead_ano);
      if (n == 2) sscanf(resultxt, "%d %d", &n, &iterations);
      if (n == 3) sscanf(resultxt, "%d %f", &n, &edge);
      if (n == 4) sscanf(resultxt, "%d %lf", &n, &limit_f);
      if (n == 5) sscanf(resultxt, "%d %lf", &n, &qual_limit);
      
      if (n == 100) sscanf(resultxt, "%d %d", &n, &o_ctrl);
      if (n == 101) sscanf(resultxt, "%d %d", &n, &f_ctrl);
   }
x_end:
   fclose(in);
   
   printf("dead_ano: %d, iterations: %d, edge: %f, limit_f: %lf, qual_limit: %lf,  o_ctrl: %d, f_ctrl: %d\n",
      dead_ano, iterations, edge, limit_f, qual_limit, o_ctrl, f_ctrl);
   
   
   pi = 3.14159;
   twopi = pi * 2.0;
   gtorad = pi / 180.0;
   
   tt = tt_old = 345.0;

   jemx = 0;		/* default: JEM-X1 (jemx = 0) */

   fillist = fopen(iname,"rt") ;	/* opening file with list of shadowgrams */
   if (fillist == NULL) {
      printf("no %s file, exiting!\n", iname);
      exit(0);
   }
   
   filnum = 0;
   
/*   printf("Opened files\n"); */

   if (o_ctrl > 0) o_ctrl = 1;
 
next_file:
   fgets(dfilnavn, 100, fillist);
   if (feof(fillist)) goto final;
   fillen = strlen(dfilnavn);
   if (filnum > 0) if (fillen < fillen_old) goto final;
   filnum++;
   fillen_old = fillen;
   
   dfilnavn[fillen-1] = 0;
   

   printf("\n\n **********************\n\nReady to read shadowgram file from %s\n", dfilnavn);

   in = fopen(dfilnavn,"rb") ;		/* opening shadowgram file, 4 shadowgrams for one JEM-X */
   if ( in == NULL) {
      printf("Error opening file: %s ..... exiting!  fillen %3d\n", dfilnavn, fillen);
      return;
   }
   
   fread(shadgram, 2, 4*detdim, in);
   fread(shdgr_aux, 8, 30, in);
   fclose(in);
   
   pha_sel[0] = shdgr_aux[17];
   pha_sel[1] = shdgr_aux[18];
   pha_sel[2] = shdgr_aux[19];
   pha_sel[3] = shdgr_aux[20];
   pha_sel[4] = shdgr_aux[21];
   jemx = shdgr_aux[22]; 

   for (i=0; i<30; i+=3) printf("shdg_aux[%2d/%2d/%2d]: %12.3lf,  %12.3lf,  %12.3lf\n", 
   i, i+1, i+2, shdgr_aux[i], shdgr_aux[i+1], shdgr_aux[i+2]);

   if (filnum == 1) {
      fprintf(cat_fil, "0  ph_boundaries (ch):        %6d   %6d   %6d   %6d   %6d\n", 
         pha_sel[0], pha_sel[1], pha_sel[2], pha_sel[3], pha_sel[4]);
      fprintf(cat_fil, "0  absorbtion length (mm):     %5.2lf  %5.2lf  %5.2lf  %5.2lf\n", 
         abs_len[0], abs_len[1], abs_len[2], abs_len[3]);
      fprintf(cat_fil, "0  x_rms scatter (mm):         %5.2lf  %5.2lf  %5.2lf  %5.2lf\n", 
         x_rms[0], x_rms[1], x_rms[2], x_rms[3]);
      fprintf(cat_fil, "0  x_rms scatter (mm):         %5.2lf  %5.2lf  %5.2lf  %5.2lf\n", 
         y_rms[0], y_rms[1], y_rms[2], y_rms[3]);
   }
   fprintf(cat_fil, "0 Shadowgram data read from %s\n", dfilnavn);

      /* shdgr_aux[4] = Orbit_number * 65536 + Pointing ID */
   aux_orbit = shdgr_aux[4] / 65536.0 + 0.01;
   aux_pid = shdgr_aux[4] - 65536.0 * aux_orbit + 0.01;
   aux_pidv = shdgr_aux[28];
   if (aux_pid < 10000) pointval = getRAdec_init(jemx, aux_orbit, aux_pid, p_RA, p_dec, p_posang);
   else  printf(" Warning: pid value exceeds array dimension!!!: %4d\n", aux_pid);

   accumT[jemx] = shdgr_aux[27];
   
   status = backsky_init(jemx, dead_ano, sky_version);
   if (status < 0) {printf("Bad status on return from backsky_init\n"); return;}


      /* ******************************************************** */
      /*                                                          */
      /*      Here starts the loop over different energy bands    */
      /*                                                          */
      /* ******************************************************** */

   for (ee=0; ee<4; ee++) {	/* ee = 0, 1, 2, 3 indicates the energy band */

	/* transfer the raw shadowgram to be analyzed into "shadow_r[]" */
      j = s_accept[ee] = s_reject[ee] = 0;
      for (i=0; i<detdim; i++) {
         soft_gr[i] = soft_shdgr[ee][i];
         if (soft_gr[i] > 0.0) {
            shadow_r[i] = shadgram[ee][i]; 
            s_accept[ee] += shadgram[ee][i];
            j += shadow_r[i]; 
         } else {
            shadow_r[i] = 0;
            s_reject[ee] += shadgram[ee][i];
         }
      }
      fprintf(cat_fil, "0  Accepted and rejected shadowgram counts: E%1d: %8d  %8d\n", 
         ee, s_accept[ee], s_reject[ee]);

      sh_rms[ee][0] = rmscalc_i(shadow_r, soft_gr);
      
	/* pixelfold generates a shadowgram "shadow" which is corrected for detector nonlinearities */   
      totcnt[jemx][ee] = pixelfold(shadow_r, pixel_st, pixel_list, shadow, ee);

      if (totcnt[jemx][ee] == 0) {
         nsrc_g[jemx][ee] = backmodels;
         continue;
      }
      
      sh_rms[ee][1] = rmscalc(shadow, soft_gr);
      
      fk = 0.0;
      for (i=0; i<detdim; i++) if (soft_gr[i] > 0.0) fk += shadow[i]; 

      printf("JEM-X%1d. Cnts before pixelfold: %6d, after: %6d, check: %6.0f, E%1d", 
         jemx+1, j, totcnt[jemx][ee], fk, ee);
         
      if (totcnt[jemx][ee] <= 0) goto next_file;

      absolen = abs_len[ee];

         /* calculate the position scatter around the X-ray absorbtion point at this energy */
      scatt_len = sh_xy(x_rms[ee], y_rms[ee], scatt_list, scatt_delta); 

      /* *********************************************************************** */
      /*                                                                         */
      /*  The "irosloop" go through the full "Iterative Removal Of Sources"      */
      /*  sequence until the loop terminates due to limited significance or      */
      /*  too many iteration loops                                               */
      /*                                                                         */
      /* *********************************************************************** */

      nsrc = irosloop(shadow, jemx, ee, xx, yy, norm_d, solut, sumfun, &goodrms[jemx][ee], &sh_rms2, 
         	&goodfits[ee], &goodnsrc[jemx][ee], resid, edge, goodbet_g[jemx][ee], iterations, limit_f);

      if (nsrc < 0) {
         n_common = n_common2 = 0;
         goto next_file;
      }
      sh_rms[ee][2] = sh_rms2;
                 
nextee:	
      /* ******************************************************** */
      /*                                                          */
      /*  Analysis of this energy band completed. Store results   */
      /*  for later comparison with the other energy bands.       */
      /*                                                          */
      /* ******************************************************** */
      
      printf("\n");
      dfilnavn[17] = 0;
      fprintf(srclistf, "%s, jemx: %1d,  E: %d, t: %6.0lf s, Const. fitpwr, %8.1lf, Background fitpwr, %8.1lf\n",
         dfilnavn, jemx+1, ee, accumT[jemx], solut[0], solut[1]);

      nsrc_g[jemx][ee] = nsrc;

      for (i=0; i<nsrc; i++) { 

         xx_g[jemx][ee][i] = xx[i];
         yy_g[jemx][ee][i] = yy[i];
         if (i >= backmodels) norm_d_g[jemx][ee][i] = 0.01 / (norm_d[i] * sumfun[i]); 

         if ((pointval > 0) && (i >= backmodels)) {
                        
            	/* get the background area corresponding to each source */
            backcnt[jemx][ee][i] = 2.0 * fun(i, 0, ee, xx, yy, jemx, norm_d, 0);
            
/*            printf("Src: %2d, E: %1d, Background estimate: %8.1f, cm2: %6.2f, total: %6d\n",
                  i, ee, backcnt[jemx][ee][i], 0.01/norm_d[i], totcnt[jemx][ee]);
*/

   	/* determine RA and dec of the source candidates found at this energy */

            getRAdec(xx[i], yy[i], jemx, &s_RA, &s_dec);
           
            fprintf(srclistf, "src: %1d, fitpwr: %8.1lf, XY: %+6.1f %+6.1f  E: %2d  RA/dec: %+7.3f %+7.3f \n", 
               i, solut[i], xx[i], yy[i], ee, s_RA, s_dec); 
            printf(
               "src: %1d, fitpwr: %8.1lf, XY: %+6.1f %+6.1f  E: %2d  RA/dec: %+7.3f %+7.3f\n", 
               i, solut[i], xx[i], yy[i], ee, s_RA, s_dec); 

               /* establish data list for all candidates found - ought to be coded as a C-structure */
               
            m_lst_ra[m_num] = s_RA;
            m_lst_dec[m_num] = s_dec;
            m_lst_pid[m_num] = aux_orbit*100000 + aux_pid * 100 + aux_pidv;
            m_lst_x[m_num] = xx[i];
            m_lst_y[m_num] = yy[i];
            m_lst_s[m_num] = solut[i] * solut[i];
            m_lst_e[m_num] = ee;
            m_lst_jx[m_num] = jemx+1;
            m_lst_ee[m_num] = ee;
            m_lst_T[m_num] = accumT[jemx];
            m_lst_s0[m_num] = solut[0] * solut[0];
            m_lst_s1[m_num] = solut[1] * solut[1];
            m_lst_T0[m_num] = shdgr_aux[23];
            m_lst_T1[m_num] = shdgr_aux[24];
            m_lst_cnt[m_num] = backcnt[jemx][ee][i]; 
            m_lst_q[m_num] = goodbet_g[jemx][ee][i] / sqrt((double)(totcnt[jemx][ee])); 

            if (i >= backmodels) m_lst_area[m_num] = norm_d_g[jemx][ee][i];
            else  m_lst_area[m_num] = 0.01 * soft_tot[26+ee];

            m_num++;

            if (m_num > 19999) m_num = 19999;
         } else {
            if (i >= backmodels) fprintf(srclistf, "src: %1d, fitpwr: %8.1lf, XY: %+6.1f %+6.1f  E: %2d\n", 
               i, solut[i], xx[i], yy[i], ee);
         }
      }

      for (i=0; i<detdim; i++) {
         shadow1[ee][i] = shadow[i];
         soft_gr1[ee][i] = soft_gr[i];
      }
   }	/* end of loop over 4 energy bands */

   for (iee=0; iee<4; iee++) {

      if (totcnt[jemx][iee] == 0.0) continue;

      printf("\nngrp[%1d]: %d, Best srcnum: %2d, best/final fitcount: %2d, best rms: %6.3f\n",
         iee, nsrc_g[jemx][iee]-backmodels, goodnsrc[jemx][iee], goodfits[iee], goodrms[jemx][iee]);
      printf("Best fit coefficients    : ");
      for (j=0; j<nsrc_g[jemx][iee]-1; j++) printf(" %7.1f", goodbet_g[jemx][iee][j]);
      printf("\n");
      printf("Corresp. eff. areas (cm2): ");
      for (j=0; j<nsrc_g[jemx][iee]-1; j++) printf(" %7.1f", norm_d_g[jemx][iee][j]);
      printf("\n");
      printf("Integration time: %8.1lf (s),  Absorbtion length: %5.2lf (mm)\n", accumT[jemx], abs_len[iee]);
   }


      /* ******************************************************** */
      /*                                                          */
      /*    print of combined results from several energy bands   */
      /*                                                          */
      /* ******************************************************** */
      
   startT = (shdgr_aux[23] - (int)(shdgr_aux[23])) * 24.0;
   starTh = startT;
   starTm = (startT - starTh) * 60;
   stopT = (shdgr_aux[24] - (int)(shdgr_aux[24])) * 24.0;
   stopTh = stopT;
   stopTm = (stopT - stopTh) * 60;

   for (i=0; i<sourcmax; i++) {
         used[jemx][0][i] = used[jemx][1][i] = used[jemx][2][i] = used[jemx][3][i] = 0;
         used2[i] = 0;
   }

   printf("\nNumber of sources: %2d %2d %2d %2d:\n",      
         nsrc_g[jemx][0]-backmodels, nsrc_g[jemx][1]-backmodels, 
         nsrc_g[jemx][2]-backmodels, nsrc_g[jemx][3]-backmodels); 
         
   k = 0;
   for (e1=0; e1<3; e1++) {
         if (nsrc_g[jemx][e1] > backmodels) { 
            for (i=backmodels; i<nsrc_g[jemx][e1]; i++) {
            
               if (used[jemx][e1][i] > 0) continue;
               
               quality = (double)(goodbet_g[jemx][e1][i]) / 
                  	sqrt((double)(totcnt[jemx][e1] * norm_d_g[jemx][e1][i] * 100.0 / soft_tot[26+e1]));

               if (quality < 1.0) continue; 
   	       	
               sumX = xx_g[jemx][e1][i] * goodbet_g[jemx][e1][i];
               sumY = yy_g[jemx][e1][i] * goodbet_g[jemx][e1][i];
               sumgood = sumsn = goodbet_g[jemx][e1][i];
               sumtot = totsn = backcnt[jemx][e1][i];
               sn_best = sumgood / sqrt((double)(sumtot));
                  
               for (e2=e1+1; e2<4; e2++) {
                  
                  if ((nsrc_g[jemx][e2]-backmodels) <= 0) continue;
                  for (j=backmodels; j<nsrc_g[jemx][e2]; j++) {
                     if (used[jemx][e2][j] > 0) continue;
                     if (goodbet_g[jemx][e2][j] < 
                  	sqrt((double)(totcnt[jemx][ee] * norm_d_g[jemx][ee][j] * 100.0 / soft_tot[26+ee]))) continue; 

                     if (sqrt((double)((xx_g[jemx][e1][i]-xx_g[jemx][e2][j])*(xx_g[jemx][e1][i]-xx_g[jemx][e2][j]) + 
                           (yy_g[jemx][e1][i]-yy_g[jemx][e2][j])*(yy_g[jemx][e1][i]-yy_g[jemx][e2][j]))) < 1.5) { 

                        tempgood = sumsn + goodbet_g[jemx][e2][j];
                        temptot = totsn + backcnt[jemx][e2][j];
                        sn_temp = tempgood / sqrt((double)(temptot));
                        sn_new = goodbet_g[jemx][e2][j] / sqrt((double)(backcnt[jemx][e2][j]));

                        if ((sn_temp >= sn_best) && (sn_temp >= sn_new )) {
                           sn_best = sn_temp;
                           sumsn = tempgood;
                           totsn = temptot;
                        }
                        if ((sn_new  >= sn_best) && (sn_new  >= sn_temp)) {
                           sn_best = sn_new;
                           sumsn = goodbet_g[jemx][e2][j];
                           totsn = totcnt[jemx][e2];
                        }
                        
                        sumgood += goodbet_g[jemx][e2][j];
                        
                        sumX += xx_g[jemx][e2][j] * goodbet_g[jemx][e2][j];
                        sumY += yy_g[jemx][e2][j] * goodbet_g[jemx][e2][j];
                        
                        avX0 = sumX / sumgood;
                        avY0 = sumY / sumgood;
                      
                        if (pointval > 0) {
                        
                           getRAdec(avX0, avY0, jemx, &s_RA, &s_dec);

                           printf("%s RA/dec: %+7.3f %+7.3f  fit: %1d %7.1lf %1d %7.1lf ",
                              dfilnavn, s_RA, s_dec, e1, goodbet_g[jemx][e1][i], e2, goodbet_g[jemx][e2][j]);
                           
                           fprintf(srclistf, "%s  RA/dec: %+7.3f %+7.3f  fit: %1d %7.1lf %1d %7.1lf ",
                              dfilnavn,  s_RA, s_dec, e1, goodbet_g[jemx][e1][i], e2, goodbet_g[jemx][e2][j]);
                           
                        } else {
                           printf("%s  XY: %+7.1f %+7.1f  strengths: %7.1lf %7.1lf ",
                              dfilnavn, avX0-127.0, avY0-127.0, goodbet_g[jemx][e1][i], goodbet_g[jemx][e2][j]);

                           fprintf(srclistf, "%s  XY: %+7.1f %+7.1f  strengths: %7.1lf %7.1lf ",
                              dfilnavn, avX0-127.0, avY0-127.0, goodbet_g[jemx][e1][i], goodbet_g[jemx][e2][j]);
                        }
                        
                        if ((starTh + starTm + stopTh + stopTm) == 0) printf("\n");
                        else printf("time: %2d:%2d to %2d:%2d\n", starTh, starTm, stopTh, stopTm);
                        
                        if ((starTh + starTm + stopTh + stopTm) == 0) fprintf(srclistf, "\n");
                        else fprintf(srclistf, "time: %2d:%2d to %2d:%2d\n", starTh, starTm, stopTh, stopTm);

                        used[jemx][e2][j] = 1;
                        used2[i] = 1;
                        accum0x[accu] = xx_g[jemx][e1][i]; 
                        accum0y[accu] = yy_g[jemx][e1][i]; 
                        accum1x[accu] = xx_g[jemx][e2][j]; 
                        accum1y[accu] = yy_g[jemx][e2][j];
                        a_orbit[accu] = aux_orbit;
                        a_pid[accu] = aux_pid;
                        a_pidv[accu] = aux_pidv;
                        accums0[accu] = goodbet_g[jemx][e1][i];
                        accums1[accu] = goodbet_g[jemx][e2][j];
                        actot0[accu] = backcnt[jemx][e1][i];
                        actot1[accu] = backcnt[jemx][e2][j];
                        accucode[accu] = jemx * 10000 + e2 * 100 + e1;
                        accumRA[accu] = s_RA;
                        accumdec[accu] = s_dec;
                        accumx[accu] = avX0;
                        accumy[accu] = avY0;
                        accums[accu] = sn_best;
                        accumn0[accu] = mnum_g[jemx][e1][i]; 
                        accumn1[accu] = mnum_g[jemx][e2][j]; 

                        com_x[0][k] = xx_g[jemx][e1][i]; 
                        com_y[0][k] = yy_g[jemx][e1][i];
                        norm_c[0][k] = norm_d_g[jemx][e1][i];
                        norm_c[1][k] = norm_d_g[jemx][e2][j];
                        com_x[1][k] = xx_g[jemx][e2][j]; 
                        com_y[1][k] = yy_g[jemx][e2][j];
                        com_s[0][k] = goodbet_g[jemx][e1][i];
                        com_s[1][k] = goodbet_g[jemx][e2][j];
                        com_avx[k] = avX0;
                        com_avy[k] = avY0;
                        com_sn[k] = sn_best;
                        com_RA[k] = s_RA;
                        com_dec[k] = s_dec;
                        com_n[0][k] = mnum_g[jemx][e1][i]; 
                        com_n[1][k] = mnum_g[jemx][e2][j]; 
                        eee[0][k] = e1;
                        eee[1][k] = e2;
                        printf("Energy scales: %d %d, n: %d, XY0: %6.1f %6.1f, XY1: %6.1f %6.1f \n", 
                           e1, e2, k, accum0x[accu], accum0y[accu], accum1x[accu], accum1y[accu]);
                     }
                  }
               }
               
               if (used2[i] > 0) {
                  used[jemx][e1][i] = 1; 
                  accu++;
                  if (accu >= accumax) accu = accumax-1;
                  k++;
               }
               used2[i] = 0;
            }
         }
   }
   fprintf(srclistf, "\n");
   printf("\n");

   n_common = k;

   for (e1=0; e1<4; e1++) {
         if (nsrc_g[jemx][e1] > backmodels) { 
            for (i=backmodels; i<nsrc_g[jemx][e1]; i++) {
            
               if (used[jemx][e1][i] > 0) continue;
               
               quality = (double)(goodbet_g[jemx][e1][i]) / sqrt((double)(totcnt[jemx][e1]));
/*               printf("Single band source:  ee: %1d,  quality: %5.1lf\n", e1, quality);
*/
               if (quality > qual_limit) {

                  printf("Single band strong source:  quality: %5.1lf  ", quality);
                  if (pointval > 0) {      
                     getRAdec(xx_g[jemx][e1][i], yy_g[jemx][e1][i], jemx, &s_RA, &s_dec);
                     printf("file: %s RA/dec: %+7.3f %+7.3f  ee: %1d  fit: %7.1lf ",
                              dfilnavn, s_RA, s_dec, e1, goodbet_g[jemx][e1][i]);
                     fprintf(srclistf, "%s  RA/dec: %+7.3f %+7.3f  ee: %1d  fit: %7.1lf",
                              dfilnavn,  s_RA, s_dec, e1, goodbet_g[jemx][e1][i]);
                  } else {
                     printf("%s  XY: %+7.1f %+7.1f  strengths: %7.1lf ",
                              dfilnavn, avX0-127.0, avY0-127.0, goodbet_g[jemx][e1][i]);
                     fprintf(srclistf, "%s  XY: %+7.1f %+7.1f  strengths: %7.1lf ",
                              dfilnavn, avX0-127.0, avY0-127.0, goodbet_g[jemx][e1][i]);
                  }
               
                  com_avx[k] = xx_g[jemx][e1][i]; 
                  com_avy[k] = yy_g[jemx][e1][i];
                  com_sn[k] = goodbet_g[jemx][e1][i] / sqrt((double)(backcnt[jemx][e1][i]));
                  com_RA[k] = s_RA;
                  com_dec[k] = s_dec;                  
                  k++;
                  fprintf(srclistf, "\n");
                  printf("\n");


                  accum0x[accu] = xx_g[jemx][e1][i]; 
                  accum0y[accu] = yy_g[jemx][e1][i]; 
                  accum1x[accu] = xx_g[jemx][e1][i]; 
                  accum1y[accu] = yy_g[jemx][e1][i];
                  a_orbit[accu] = aux_orbit;
                  a_pid[accu] = aux_pid;
                  a_pidv[accu] = aux_pidv;
                  accums0[accu] = goodbet_g[jemx][e1][i];
                  accums1[accu] = goodbet_g[jemx][e1][i];
                  actot0[accu] = backcnt[jemx][e1][i];
                  actot1[accu] = backcnt[jemx][e1][i];
                  accucode[accu] = jemx * 10000 + e1 * 100 + e1;
                  accumRA[accu] = s_RA;
                  accumdec[accu] = s_dec;
                  accumx[accu] = xx_g[jemx][e1][i]; 
                  accumy[accu] = yy_g[jemx][e1][i]; 
                  accums[accu] = goodbet_g[jemx][e1][i] / sqrt((double)(backcnt[jemx][e1][i]));
                  accumn0[accu] = mnum_g[jemx][e1][i]; 
                  accumn1[accu] = mnum_g[jemx][e1][i]; 
                  accu++;
               }
            }
         }
   }
                  
   n_common2 = k;
      
      /* ******************************************************** */
      /*                                                          */
      /*      redo source fit using only sources detected        */
      /*            in more than one energy band                  */
      /*                                                          */
      /* ******************************************************** */
      
      if (n_common2 > n_common)
         printf("\n Analysis using %d common and %d strong source(s) between 4 energy bands\n", 
            n_common, n_common2-n_common);
      else
         printf("\n Analysis using %d common source(s) between 4 energy bands\n", n_common);

      ksrc = backmodels + n_common2;

      for (i=0; i<n_common2; i++) {
         xx[backmodels+i] = com_avx[i];
         yy[backmodels+i] = com_avy[i];

         chk_no = 1;
         printf("source %1d:  X: %6.2f, Y: %6.2f\n", i, xx[backmodels+i], yy[backmodels+i]);

      }
         
      for (ee=0; ee<4; ee++) {
         
         if (n_common == 0) {printf("No common sources in the four energy bands\n"); goto w_fits;}

         xrms = x_rms[ee];
         yrms = y_rms[ee];

         for (i=0; i<skydim; i++) imgpeak[ee][i] = 0.0;
      
         for (i=0; i<detdim; i++) {
            shadow[i] = shadow1[ee][i];
            soft_gr[i] = soft_gr1[ee][i];
         }
   
         absolen = abs_len[ee];
         scatt_len = sh_xy(xrms, yrms, scatt_list, scatt_delta); 
            /* calculate read-out scatter around X-ray absorbtion point at this energy */

         fun(-1, 1, ee, xx, yy, jemx, norm_d, 0); 	/* reset fitting functions */

         finalvar[ee] = imagefit(shadow, jemx, ee, resid, ksrc, xx, yy, solut, norm_d, sumfun, chk_no);
   
         sh_rms[ee][3] = rmscalc(resid, soft_gr);

         /* ************************************************************** */
         /*                                                                */
         /*    Backproject final residual shadowgram to form sky image     */
         /*                                                                */
         /* ************************************************************** */
      
         backpro_d(resid, skyim_r);
         clean2(skyim_r, xx, yy, n_common, backmodels, imgpeak[ee]);

         for (i=0; i<n_common2; i++) { 
            peaktransfer(xx[backmodels+i], yy[backmodels+i], fitsimage[ee][2], imgpeak[ee]);
         }
         
         /* redo cleaning of raw image, now using the information about strong sources */
         clean2(fitsimage[ee][1], xx, yy, n_common2, backmodels, fitsimage[ee][2]);
         /* recover peaks of strong sources from 'imgpeak' image */
         for (i=0; i<n_common2; i++) { 
            peaktransfer(xx[backmodels+i], yy[backmodels+i], imgpeak[ee], fitsimage[ee][2]);
         }
         
         
            /* after final fit output final, source cleaned map */
w_fits:             
         sprintf(fitsnavn, "skyima%1d_%1d_%04d_%03d%1d.fits", jemx+1, ee, aux_orbit, aux_pid, aux_pidv);

         bitpix = -32;		/* use float representation in FITS file */
         dim1 = dim2 = FOV_radius * 2 + 1;	/* midisky image dimensions */
     
         cdelt = (mask_r + coll_r) * 180.0 / (FOV_radius * mask_height * pi);
/*         if (jemx == 0) jmxsign = 1; else jmxsign = -1;
*/
   	jmxsign = -1;
         rad_ang = p_posang[aux_pid] * pi / 180.0 + pi / 2.0 + phase2;
         printf("At Funopen: cdelt: %10.8lf, rad_ang: %10.6lf, phase2: %8.6lf\n", cdelt, rad_ang, phase2);

         if (f_ctrl == 0) goto noFun;
         
         fi_headr(fitsnavn);

         if( !(fun_handle = FunOpen(fitsnavn, "a", NULL)) ) 
            gerror(stderr, "could not FunOpen output file: %s\n", fitsnavn);
         else {
            /* prepare header */
            
            FunParamPuts(fun_handle, "EXTNAME", 0, "RAW_RECT", "Rectified initial intensity map", 1);
            FunParamPuts(fun_handle, "IMATYPE", 0, "INTENSITY", "Raw intensity map", 1);
            FunParamPuti(fun_handle, "VARIA_ID", 0, 100, "Variance map identifier", 1);
            FunParamPuts(fun_handle, "CTYPE1", 0, "RA---TAN", NULL, 1);
            FunParamPuts(fun_handle, "CTYPE2", 0, "DEC--TAN", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT1", 0, "deg", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT2", 0, "deg", NULL, 1);
            if (jemx == 1) {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx2", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-2.txt", "Vignetting map name", 1);
            } else {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx1", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-1.txt", "Vignetting map name", 1);
            }
            FunParamPutd(fun_handle, "CRVAL1", 0, p_RA[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRVAL2", 0, p_dec[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "ROT_ANG", 0, p_posang[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX1", 0, (double)(128.0 + jmxsign*offY), 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX2", 0, (double)(128.0 + jmxsign*offX), 12, NULL, 1);
 
            FunParamPutd(fun_handle, "CD1_1", 0, cdelt * cos(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD1_2", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_1", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_2", 0, -cdelt * cos(rad_ang), 12, NULL, 1);
 
            FunParamPutd(fun_handle, "EXPOSURE", 0, accumT[jemx], 12, "Effective integration time (s)", 1);
            FunParamPuti(fun_handle, "JMX_UNIT", 0, jemx+1, NULL, 1);
            FunParamPuti(fun_handle, "REVOL", 0, aux_orbit, NULL, 1);
            FunParamPuti(fun_handle, "PID", 0, aux_pid, NULL, 1);
            FunParamPuti(fun_handle, "PIDV", 0, aux_pidv, NULL, 1);
            FunParamPutd(fun_handle, "E_MIN", 0, eval[ee], 12, "keV", 1);
            FunParamPutd(fun_handle, "E_MAX", 0, eval[ee+1], 12, "keV", 1);

            if( !FunImagePut(fun_handle, fitsimage[ee][2], dim1, dim2, bitpix, NULL) ) 
               gerror(stderr, "could not FunImagePut: %s\n", fitsnavn);

            FunClose(fun_handle);
         }
         
         if( !(fun_handle = FunOpen(fitsnavn, "a", NULL)) ) 
            gerror(stderr, "could not FunOpen output file: %s\n", fitsnavn);
         else {

            FunParamPuts(fun_handle, "EXTNAME", 0, "VAR", "Variance map", 1);
            FunParamPuts(fun_handle, "IMATYPE", 0, "VARIANCE", "Variance map", 1);
            FunParamPuti(fun_handle, "VARIA_ID", 0, 100, "Variance map identifier", 1);
            FunParamPuts(fun_handle, "CTYPE1", 0, "RA---TAN", NULL, 1);
            FunParamPuts(fun_handle, "CTYPE2", 0, "DEC--TAN", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT1", 0, "deg", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT2", 0, "deg", NULL, 1);
            if (jemx == 1) {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx2", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-2.txt", "Vignetting map name", 1);
            } else {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx1", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-1.txt", "Vignetting map name", 1);
            }
            FunParamPutd(fun_handle, "CRVAL1", 0, p_RA[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRVAL2", 0, p_dec[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "ROT_ANG", 0, p_posang[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX1", 0, (double)(128.0 + jmxsign*offY), 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX2", 0, (double)(128.0 + jmxsign*offX), 12, NULL, 1);

            FunParamPutd(fun_handle, "CD1_1", 0, cdelt * cos(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD1_2", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_1", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_2", 0, -cdelt * cos(rad_ang), 12, NULL, 1);

            FunParamPutd(fun_handle, "EXPOSURE", 0, accumT[jemx], 12, "Effective integration time (s)", 1);
            FunParamPuti(fun_handle, "JMX_UNIT", 0, jemx+1, NULL, 1);
            FunParamPuti(fun_handle, "REVOL", 0, aux_orbit, NULL, 1);
            FunParamPuti(fun_handle, "PID", 0, aux_pid, NULL, 1);
            FunParamPuti(fun_handle, "PIDV", 0, aux_pidv, NULL, 1);
            FunParamPutd(fun_handle, "E_MIN", 0, eval[ee], 12, "keV", 1);
            FunParamPutd(fun_handle, "E_MAX", 0, eval[ee+1], 12, "keV", 1);

            if( !FunImagePut(fun_handle, fitsimage[ee][0], dim1, dim2, bitpix, NULL) ) 
               gerror(stderr, "could not FunImagePut: %s\n", fitsnavn);

            FunClose(fun_handle);
         }

         if (n_common2 == 0) continue;	/* goto next ee-value */
         
         if( !(fun_handle = FunOpen(fitsnavn, "a", NULL)) ) 
            gerror(stderr, "could not FunOpen output file: %s\n", fitsnavn);
         else {
            FunParamPuts(fun_handle, "EXTNAME", 0, "RESIDUAL", "Residual intensity map - no flat fielding", 1);
            FunParamPuts(fun_handle, "IMATYPE", 0, "RESIDUAL", "Residual map (detected sources subtracted)", 1);
            FunParamPuti(fun_handle, "VARIA_ID", 0, 100, "Variance map identifier", 1);
            FunParamPuts(fun_handle, "CTYPE1", 0, "RA---TAN", NULL, 1);
            FunParamPuts(fun_handle, "CTYPE2", 0, "DEC--TAN", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT1", 0, "deg", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT2", 0, "deg", NULL, 1);
            if (jemx == 1) {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx2", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-2.txt", "Vignetting map name", 1);
            } else {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx1", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-1.txt", "Vignetting map name", 1);
            }
            FunParamPutd(fun_handle, "CRVAL1", 0, p_RA[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRVAL2", 0, p_dec[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "ROT_ANG", 0, p_posang[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX1", 0, (double)(128.0 + jmxsign*offY), 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX2", 0, (double)(128.0 + jmxsign*offX), 12, NULL, 1);

            FunParamPutd(fun_handle, "CD1_1", 0, cdelt * cos(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD1_2", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_1", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_2", 0, -cdelt * cos(rad_ang), 12, NULL, 1);

            FunParamPutd(fun_handle, "EXPOSURE", 0, accumT[jemx], 12, "Effective integration time (s)", 1);
            FunParamPuti(fun_handle, "JMX_UNIT", 0, jemx+1, NULL, 1);
            FunParamPuti(fun_handle, "REVOL", 0, aux_orbit, NULL, 1);
            FunParamPuti(fun_handle, "PID", 0, aux_pid, NULL, 1);
            FunParamPuti(fun_handle, "PIDV", 0, aux_pidv, NULL, 1);
            FunParamPutd(fun_handle, "E_MIN", 0, eval[ee], 12, "keV", 1);
            FunParamPutd(fun_handle, "E_MAX", 0, eval[ee+1], 12, "keV", 1);
            FunParamPuti(fun_handle, "N_SRC", 0, n_common2, "Number of sources extracted", 1);

            if( !FunImagePut(fun_handle, skyim_r, dim1, dim2, bitpix, NULL) ) 
               gerror(stderr, "could not FunImagePut: %s\n", fitsnavn);

            FunClose(fun_handle);
         }

         if( !(fun_handle = FunOpen(fitsnavn, "a", NULL)) ) 
            gerror(stderr, "could not FunOpen output file: %s\n", fitsnavn);
         else {

            FunParamPuts(fun_handle, "EXTNAME", 0, "RES+SRC", "Residual intensity + sources", 1);
            FunParamPuts(fun_handle, "IMATYPE", 0, "RES+SRC", "Residual map + brushed up with sources", 1);
            FunParamPuti(fun_handle, "VARIA_ID", 0, 100, "Variance map identifier", 1);
            FunParamPuts(fun_handle, "CTYPE1", 0, "RA---TAN", NULL, 1);
            FunParamPuts(fun_handle, "CTYPE2", 0, "DEC--TAN", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT1", 0, "deg", NULL, 1);
            FunParamPuts(fun_handle, "CUNIT2", 0, "deg", NULL, 1);
            if (jemx == 1) {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx2", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-2.txt", "Vignetting map name", 1);
            } else {
               FunParamPuts(fun_handle, "INSTRUME", 0, "jmx1", NULL, 1);
               FunParamPuts(fun_handle, "VIGNDOL", 0, "illum_l-1.txt", "Vignetting map name", 1);
            }
            FunParamPutd(fun_handle, "CRVAL1", 0, p_RA[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRVAL2", 0, p_dec[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "ROT_ANG", 0, p_posang[aux_pid], 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX1", 0, (double)(128.0 + jmxsign*offY), 12, NULL, 1);
            FunParamPutd(fun_handle, "CRPIX2", 0, (double)(128.0 + jmxsign*offX), 12, NULL, 1);

            FunParamPutd(fun_handle, "CD1_1", 0, cdelt * cos(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD1_2", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_1", 0, cdelt * sin(rad_ang), 12, NULL, 1);
            FunParamPutd(fun_handle, "CD2_2", 0, -cdelt * cos(rad_ang), 12, NULL, 1);

            FunParamPutd(fun_handle, "EXPOSURE", 0, accumT[jemx], 12, "Effective integration time (s)", 1);
            FunParamPuti(fun_handle, "JMX_UNIT", 0, jemx+1, NULL, 1);
            FunParamPuti(fun_handle, "REVOL", 0, aux_orbit, NULL, 1);
            FunParamPuti(fun_handle, "PID", 0, aux_pid, NULL, 1);
            FunParamPuti(fun_handle, "PIDV", 0, aux_pidv, NULL, 1);
            FunParamPutd(fun_handle, "E_MIN", 0, eval[ee], 12, "keV", 1);
            FunParamPutd(fun_handle, "E_MAX", 0, eval[ee+1], 12, "keV", 1);
            FunParamPuti(fun_handle, "N_SRC", 0, n_common2, "Number of sources extracted", 1);
            FunParamPuts(fun_handle, "COMMENT", 0, NULL, "This image contains the residual intensity", 1);
            FunParamPuts(fun_handle, "COMMENT", 0, NULL, "map with the raw source peaks restored", 1);
            FunParamPuts(fun_handle, "COMMENT", 0, NULL, "Only the pixels within the selected peak", 1); 
            FunParamPuts(fun_handle, "COMMENT", 0, NULL, "areas (5x5 pixels) have been transferred",1);

            if( !FunImagePut(fun_handle, imgpeak[ee], dim1, dim2, bitpix, NULL) ) 
               gerror(stderr, "could not FunImagePut: %s\n", fitsnavn);

            FunClose(fun_handle);
         }
noFun:
         if (j > 0) if (fabs(solut[0]) > solutold) goto loopstop4;

         solutold = fabs(solut[0]);
         rms_old = rms;

loopstop4:
         for (j=0; j<backmodels+n_common2; j++) {
            if (j >= backmodels) norm_d_g[jemx][ee][j] = 0.01 / (norm_d[j] * sumfun[j]); 
            else norm_d_g[jemx][ee][j] = 0.01 * soft_tot[26+ee];
            solut_e[ee][j] = solut[j];
         }
         
         if (norm_d_g[jemx][ee][0] > 0.0) {
            finalvar[ee] /= totcnt[jemx][ee] * 0.01 / norm_d_g[jemx][ee][0];	/* form reduced variance */
         
            printf(
               "E%1d shadowgram rel_sigma: raw: %10.3lf, pix_f: %10.3lf, backgr_fit: %10.3lf, src_fit: %10.3lf\n",
               ee, sh_rms[ee][0], sh_rms[ee][1], sh_rms[ee][2], sh_rms[ee][3]);
         }
/*
fprintf(srclist2, "E%1d shadowgram rel_sigma: raw: %10.3lf, pix_f: %10.3lf, backgr_fit: %10.3lf, src_fit: %10.3lf\n",
            ee, sh_rms[ee][0], sh_rms[ee][1], sh_rms[ee][2], sh_rms[ee][3]);
*/
         if (ee < 3) continue;
         
         dfilnavn[12] = 0;

         sprintf(resultxt, 
            "\n%s   Orbit: %3d, PID(scw): %3d %d, JEM-X%1d,  No of sources: %2d, Integr. time: %4.0lf s, ",
            dfilnavn, aux_orbit, aux_pid, aux_pidv, jemx+1, n_common2, accumT[jemx]);
         sprintf(dummy, "Pointing: RA/dec/R: %7.3lf %+7.3lf %6.2lf\n", 
            p_RA[aux_pid], p_dec[aux_pid], p_posang[aux_pid]);

         strcat(resultxt, dummy);
         printf("%s", resultxt);
         fprintf(srclist2, "%s", resultxt);

         sprintf(resultxt, 
            "Total cnts %5d %5d %5d %5d,   Constant cps: %5.2f %5.2f %5.2f %5.2f, ",
            totcnt[jemx][0], totcnt[jemx][1], totcnt[jemx][2], totcnt[jemx][3],  
            (float)(solut_e[0][0] / accumT[jemx]), (float)(solut_e[1][0] / accumT[jemx]), 
            (float)(solut_e[2][0] / accumT[jemx]), (float)(solut_e[3][0] / accumT[jemx])); 
         
         sprintf(dummy, " Backgr. cps: %5.2f %5.2f %5.2f %5.2f\n",
            (float)(solut_e[0][1] / accumT[jemx]), (float)(solut_e[1][1] / accumT[jemx]), 
            (float)(solut_e[2][1] / accumT[jemx]), (float)(solut_e[3][1] / accumT[jemx])); 

         strcat(resultxt, dummy);
         printf("%s", resultxt);
         fprintf(srclist2, "%s", resultxt);

         for (j=0; j<n_common2; j++) {
            jb = backmodels + j;
            
            cmx = com_avx[j] - FOV_radius;
            cmy = com_avy[j] - FOV_radius;
            avr = sqrt(cmx*cmx + cmy*cmy);
            
            sprintf(resultxt, "x/y/r: %6.2f %6.2f %6.2f, RA/dec: %6.2f %+6.2f, SN: %6.2f, ",
            	cmx, cmy, avr, com_RA[j], com_dec[j], com_sn[j]);

            sprintf(dummy, "Real/Norm cps: %6.3f %6.2f  %6.3f %6.2f  %6.3f %6.2f  %6.3f %6.2f\n",
                (float)(solut_e[0][jb] / accumT[jemx]), 
            	(float)(solut_e[0][jb] * 100.0 / (accumT[jemx] * norm_d_g[jemx][0][jb])), 
                (float)(solut_e[1][jb] / accumT[jemx]), 
                (float)(solut_e[1][jb] * 100.0 / (accumT[jemx] * norm_d_g[jemx][1][jb])), 
                (float)(solut_e[2][jb] / accumT[jemx]), 
                (float)(solut_e[2][jb] * 100.0 / (accumT[jemx] * norm_d_g[jemx][2][jb])), 
                (float)(solut_e[3][jb] / accumT[jemx]), 
                (float)(solut_e[3][jb] * 100.0 / (accumT[jemx] * norm_d_g[jemx][3][jb])));

            strcat(resultxt, dummy);
            printf("%s", resultxt);               
            fprintf(srclist2, "%s", resultxt);
         }
      }

      goto next_file;
final:

      for (i=0; i<m_num; i++) {
         used2[i] = 0;
      }
      printf("Total number of source candidates found: %5d:\n", m_num);
         
      accu2 = 0;
      accu3 = 0;
      for (i=0; i<m_num-1; i++) {
         if ((used2[i] > 0) || (m_lst_s[i] < 0.0)) continue;
         sumX = m_lst_ra[i] * m_lst_s[i];
         sumY = m_lst_dec[i] * m_lst_s[i];
         sumgood = sumsn = m_lst_s[i];
         sumcnts = totsn = m_lst_cnt[i];
         sn_best = m_lst_s[i] / sqrt((double)(m_lst_cnt[i]));
         cosdec = cos((double)(m_lst_dec[i]) * gtorad);
         sumtot = 1;
         pidlast = m_lst_pid[i];
         e_indic = 1 << (m_lst_e[i] * 7);
                  
         for (j=i+1; j<m_num; j++) {
            if ((used2[j] > 0) || (m_lst_s[j] < 0.0)) continue;
            if ((m_lst_pid[i] == m_lst_pid[j]) && (m_lst_e[i] == m_lst_e[j])) continue;
            
            if (sqrt((double)((m_lst_ra[i] -m_lst_ra[j] ) * (m_lst_ra[i] -m_lst_ra[j] ))*cosdec*cosdec + 
                     (double)((m_lst_dec[i]-m_lst_dec[j]) * (m_lst_dec[i]-m_lst_dec[j]))) < 0.05) { 

               if (used1[i] == 0) {      
                  accu2RA[accu2] = m_lst_ra[i];
                  accu2dec[accu2] = m_lst_dec[i];
                  accum2x[accu2] = m_lst_x[i];
                  accum2y[accu2] = m_lst_y[i];
                  acc2pid[accu2] = m_lst_pid[i];
                  acc2sn[accu2] =  sqrt((double)(m_lst_s[i] / m_lst_cnt[i]));
                  accums2[accu2] =  sqrt((double)(m_lst_s[i]));
                  actot2[accu2] =  m_lst_cnt[i];
                  acc2ee[accu2] = m_lst_e[i];
                  acc2jx[accu2] = jemx;
                  accu2++;
                  if (accu2 >= accumax) accu2 = accumax-1;
               }
               accu2RA[accu2] = m_lst_ra[j];
               accu2dec[accu2] = m_lst_dec[j];
               accum2x[accu2] = m_lst_x[j];
               accum2y[accu2] = m_lst_y[j];
               acc2pid[accu2] = m_lst_pid[j];
               acc2sn[accu2] =  sqrt((double)(m_lst_s[j] / m_lst_cnt[j]));
               accums2[accu2] =  sqrt((double)(m_lst_s[j]));
               actot2[accu2] =  m_lst_cnt[j];
               acc2ee[accu2] = m_lst_e[j];
               acc2jx[accu2] = jemx;
               accu2++;
               if (accu2 >= accumax) accu2 = accumax-1;

               sumX += m_lst_ra[j] * m_lst_s[j];
               sumY += m_lst_dec[j] * m_lst_s[j];

               tempgood = sumsn + m_lst_s[j];
               temptot = totsn + m_lst_cnt[j];
               sn_temp = tempgood / sqrt((double)(temptot));
               sn_new = m_lst_s[j] / sqrt((double)(m_lst_cnt[j]));

               if ((sn_temp >= sn_best) && (sn_temp >= sn_new )) {
                  sn_best = sn_temp;
                  sumsn = tempgood;
                  totsn = temptot;
               }
               if ((sn_new  >= sn_best) && (sn_new  >= sn_temp)) {
                  sn_best = sn_new;
                  sumsn = goodbet_g[jemx][e2][j];
                  totsn = totcnt[jemx][e2];
               }
                        
               sumgood += m_lst_s[j];
               sumcnts += m_lst_cnt[j];
               sumtot++;

               e_indic += 1 << (m_lst_e[j] * 7);
               
               if (pidlast != m_lst_pid[j]) sumtot += 1000; 
               pidlast = m_lst_pid[j];
                                    
               avX0 = sumX / sumgood;
               avY0 = sumY / sumgood;
               used1[i] = 1;
               used2[j] = 1;
            }
            if (used1[i] == 1) {      
               if (i == last_i) continue;
               last_i = i;
               accum3x[accu3] = avX0;
               accum3y[accu3] = avY0;
               acc3tot[accu3] = totsn;
               acc3sn[accu3] = sn_best;
               acc3ee[accu3] = e_indic;
               accu3++;
               if (accu3 >= accumax) accu3 = accumax-1;
            }
         }
      }   


   fprintf(cat_fil, "\n\n");
   for (i=0; i<accu; i++) {
      j = accucode[i] / 10000;
      k = accucode[i] % 100;
      fprintf(cat_fil, "1  %3d %4d %4d %3d %3d %3d %7.3f %7.3f %10.1f %10.1f %7.3f %7.3f \n", 
      	i, a_orbit[i], a_pid[i], a_pidv[i], j, k, accum0x[i], accum0y[i], 
        accums0[i], actot0[i], accumRA[i], accumdec[i]);
   }
   fprintf(cat_fil, "\n\n");
   for (i=0; i<accu; i++) {
      j = accucode[i] / 10000;
      k = ((accucode[i] - j * 10000) / 100) % 100;
      fprintf(cat_fil, "2  %3d %4d %4d %3d %3d %3d %7.3f %7.3f %10.1f %10.1f %7.3f %7.3f \n", 
        i, a_orbit[i], a_pid[i], a_pidv[i], j, k, accum1x[i], accum1y[i], 
        accums1[i], actot1[i], accumRA[i], accumdec[i]);
   }
   fprintf(cat_fil, "\n\n");
   
   for (i=0; i<accu3; i++) {
      ee0 = acc3ee[i] % 128;
      ee1 = (acc3ee[i] >> 7) % 128;
      ee2 = (acc3ee[i] >> 14) % 128;
      ee3 = (acc3ee[i] >> 21) % 128;
      fprintf(cat_fil, "3  %3d  %7.3f  %7.3f  %12.1f  %8d   %3d %3d %3d %3d\n", 
      i, accum3x[i], accum3y[i], acc3sn[i], acc3tot[i], ee0, ee1, ee2, ee3);
   }
   fprintf(cat_fil, "\n\n");

   for (i=0; i<accu2; i++) {
      orbit = acc2pid[i] / 100000;
      pid  = (acc2pid[i] / 100) % 1000;
      pidv  = acc2pid[i] % 100;

      fprintf(cat_fil, "4  %3d %4d %4d %3d %3d %3d", i, orbit, pid, pidv, acc2jx[i]+1, acc2ee[i]);
      fprintf(cat_fil, " %7.3f %7.3f %10.1f %10.1f", accum2x[i], accum2y[i], accums2[i], actot2[i]);
      fprintf(cat_fil, " %7.3lf %7.3lf %7.3lf %7.3f %7.3f\n", 
         		p_RA[pid], p_dec[pid], p_posang[pid], accu2RA[i], accu2dec[i]);
   }
   
   fprintf(cat_fil, "\n\n");

   fclose(cat_fil);
   fclose(srclistf);

   fclose(srclist2);
   printf("  Accepted and rejected shadowgram counts:\n");
   printf("E0:  %8d    %8d\n", s_accept[0], s_reject[0]);
   printf("E1:  %8d    %8d\n", s_accept[1], s_reject[1]);
   printf("E2:  %8d    %8d\n", s_accept[2], s_reject[2]);
   printf("E3:  %8d    %8d\n", s_accept[3], s_reject[3]);

   return;
}

/* ******************************************************************** */

int backsky_init(int jemx, int dead_ano, char sky_version)	/* initialize global parameters */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*                                                             */        
/*                                                             */        
/*                                                             */
/*  Niels Lund, February 2003                                  */        
/*                                                             */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

{
   FILE *sp_fil, *datafil, *ifil, *pofil;
   int i, j, k, m, n, bytes, indent = 1, minsum, x, y, r2, pixindex;
   char phline[128], skymapa[128], skymap[128], skyfil[200], irosfil[200], pixfil[200], skofil[200], srcnam[128];
   static int done_data_input = 0;
   double softsum;
   char dummy[100], pointfil[200];
   float indat[2];

   
      /* ++++++++++++++ The following part is only executed once ! +++++++++++++++++ */
      
   if (done_data_input == 1) goto skip; 
    
   done_data_input = 1;
   
   sprintf(skyfil, "%s/reference/skmap%1cc%1d_a", sweethome, sky_version, jemx+1); /* default backprojection file */
   printf("Default backprojection file: %s\n", skyfil);
   fprintf(cat_fil, "0  backprojection file: %s\n", skyfil);
   
   strcpy(skymapa, skyfil);
   strcpy(skymap, skyfil);
   strcat(skymapa, ".aux");
   strcat(skymap, ".dat");

/*   printf("Backprojection files: %s  %s\n", skymap, skymapa);
*/   
   datafil = fopen(skymapa, "rb");	/* opening backprojection auxiliary file */
   if (datafil == NULL) {
      printf("\nCould not open %s, goodbye\n", skymapa);
      return(-1);
   }
   
   fread(parameters, 8, 20, datafil);
   fread(d_allow_st, 4, detdim, datafil);
   fread(d_allow_l, 2, detdim, datafil);
   fread(d_allow_p, 4, detdim, datafil);
   fread(maskhole, 2, 22501, datafil);
   fread(notes, 1, 1000, datafil);

   fclose(datafil);
   
   if (dead_ano == 1) 
      sprintf(skofil, "%s/reference/period5new/smooth%1d+anod.bin", sweethome, jemx+1);
   else sprintf(skofil, "%s/reference/period5new/smooth%1d-anod.bin", sweethome, jemx+1);

   fprintf(cat_fil, "0 background reference shadowgram read from: %s\n", skofil);

   datafil = fopen(skofil, "rb");

   fread(soft_shdgr, 4, 4*detdim, datafil);
   fread(soft_tot, 8, 30, datafil);
   fclose(datafil);
   
/*   for (i=0; i<30; i++) printf("Soft_tot[%2d]: %10.4lf\n", i, soft_tot[i]);
*/   
   sprintf(pixfil, "%s/reference/pixdata%1d.bin", sweethome, jemx+1);
   fprintf(cat_fil, "0  pixel file: %s\n", pixfil);
   
   ifil = fopen(pixfil, "rb");
   if (ifil == NULL) {printf(" Error in opening %s\n", pixfil); return;}
   fread(pixel_st, 4, detdim, ifil);
   pixindex = pixel_st[0];
   fread(pixel_list, 1, pixindex, ifil);
   fclose(ifil);

   softsum = 0.0;
   for (i=1; i<22501; i++) {
      softsum += maskhole[i];
   }
        
     FOV_radius = -1;
     sprintf(pointfil, "%s/reference/alignment.dat", sweethome);
      pofil = fopen(pointfil, "rt");
      if (pofil == NULL) {
         printf("No alignment file found! (%s)\n", pointfil);
         exit(0);
      } else {
         for (i=0; i<200; i++) {
            fgets(dummy, 100, pofil);
            if (feof(pofil)) goto q_end;
            sscanf(dummy, "%d", &n);
            if (n == 4)   {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); coll_r = indat[jemx];}
            if (n == 5)   {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); mask_r = indat[jemx];}
            if (n == 6)   {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); mask_height = indat[jemx];}

            if (n == 13)  {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); xshi = indat[jemx];}
            if (n == 14)  {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); yshi = indat[jemx];}

            if (n == 20)  {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); indep = indat[jemx];}
            if (n == 21)  {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); FOV_radius = indat[jemx];}

            if (n == 100) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); scaleX = indat[jemx];}
            if (n == 101) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); scaleY = indat[jemx];}
            if (n == 102) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); phase2 = indat[jemx];}
            if (n == 103) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); x2x = indat[jemx];}
            if (n == 104) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); y2y = indat[jemx];}
            if (n == 105) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offX = indat[jemx];}
            if (n == 106) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offY = indat[jemx];}
            if (n == 107) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offQ = indat[jemx];}
            if (n == 108) {sscanf(dummy, "%d %f %f", &n, &indat[0], &indat[1]); offR = indat[jemx];}
         }
      }
q_end:
   if (FOV_radius < 0) {printf("alignment data not updated to contain FOV_radius, exiting!\n"); exit(0);}
   depth_corr = mask_height / (mask_height + indep);
   
   bytes = parameters[0];        /* length of 'd_allow' list */

   if (bytes > max_sky_bytes) {
      printf("\nd_allow underdimensioned, bytes = %10d", bytes);
      return(-1);
   }

   datafil = fopen(skymap, "rb");	/* opening backprojection data file */

   if (datafil == NULL) {
      printf("\nCould not open %s, goodbye\n", skymap);
      return(-1);
   }

   fread(d_allow, 1, bytes, datafil);

   fclose(datafil);

      /* dead_ano = 0;  default: reject bad anodes */
   if (dead_ano == 0) sprintf(irosfil,"%s/reference/detmap%1c-%1d.aux", sweethome, sky_version, jemx+1);
   else sprintf(irosfil,"%s/reference/detmap%1c+%1d.aux", sweethome, sky_version, jemx+1);
   
   datafil = fopen(irosfil, "rb");
   if (datafil == NULL) {
      printf("\nCould not open %s, goodbye\n", irosfil);
      return(-1);
   }
   fread(det_e, 4, skydim, datafil);
   fclose(datafil);
   
   m = n = 0;
   for (j=1; j<skydim; j++) {
      if (det_e[j] == 0) continue;
      if (det_e[j] > n) n = det_e[j];
      rsignif[j] = 1.0 / sqrt((double)(det_e[j]));
      if (det_e[j] > signiflim) access[j] = 1; else access[j] = 0;
      x = (float)(j / sky_ydim) - 127.5;
      y = (float)(j % sky_ydim) - 127.5;
      if ((x*x + y*y) > 16129) access[j] = 0;
      m++;
      k = det_e[j] / 50;
      if (k > 29) k = 29;
   }
   for (i=0; i<32; i++) bb[i] = 1<<i;

   sprintf(srcnam, "n_sources.txt");
   srclistf = fopen(srcnam,"wt") ;
   if ( srclistf == NULL) {
      printf("Error opening file: %s ..... exiting!\n", srcnam);
      return;
   }
   
   sprintf(srcnam, "n_sources2.txt");
   srclist2 = fopen(srcnam,"wt") ;
   if ( srclist2 == NULL) {
      printf("Error opening file: %s ..... exiting!\n", srcnam);
      return;
   }
   
skip:
   return(0);
}

/* ******************************************************************** */

float fturn(int jemx, float pos)
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*                                                             */        
/*                                                             */        
/*                                                             */
/*  Niels Lund, February 2003                                  */        
/*                                                             */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

{
   /* return pos value unchanged for JEM-X1,
      return "mirror value of pos around [skydim-1] for JEM-X2 
      (the routine assumes that pos is in the range 0 to skydim-1 */
       
   if (jemx == 0) return(pos); else return((float)(skymirror) - pos);
}

/* **************************************************************** */

int pixelfold(short det[], int pixel_st[], unsigned char pixel_list[], float det_out[], int ee)
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/*                                                             */        
/*  pixelfold.c corrects for the nonuniform readout grid of    */
/*  real microstrid detector. The counts recorded in a         */
/*  particular telemetry pixel is distributed among idealized  */
/*  1x1 mm2 pixels on an idealized microstrip detector         */        
/*                                                             */
/*  Niels Lund, February 2003                                  */        
/*                                                             */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

{
   int i, j, k, m, n, len, index, delta, val, valnum;
   double sum1, sum2, loss1, valsum, vals;

   for (i=0; i<detdim; i++) det_out[i] = 0.0;
   valsum = vals = 0.0;
   valnum = 0;
   
   sum1 = loss1 = 0.0;
   for (i=1; i<detdim; i++) {
      if ((det[i] <= 0) || (soft_gr[i] <= 0.0)) continue;
      sum1 += det[i];
      vals = 0.0;
      index = pixel_st[i];
      len = pixel_list[index++];
      for (j=0; j<len; j++) {
         delta = pixel_list[index++];
         val = pixel_list[index++];
         vals += val;
         k = ((delta & 15) - 8); 
         m = ((delta & 240) - 128) * 16;
         n = i + k + m;
         if ((n < 0) || (n >= detdim)) 
            loss1 += val * (float)(det[i]);
         else
            det_out[n] += val * (float)(det[i]);         
      }
      valsum += vals / 255.0;
      valnum++;
      if ((vals > 260.0) || (vals < 250.0)) {
         index = pixel_st[i];
         len = pixel_list[index++];
         /* printf("%6d pixel: len: %1d, [delta/delta'/val]: ", i, len); */
         n = 0;
         for (j=0; j<len; j++) {
            delta = pixel_list[index++];
            val = pixel_list[index++];
            n += val;
            k = ((delta & 15) - 8); 
            m = ((delta & 240) - 128) * 16;
            m += k;
            /* printf(" %2x/%+4d/%3d  ", delta, m, val); */
         }
         /* printf(" sum: %3d\n", n); */
         if ((vals > 260.0) || (vals < 250.0)) exit(0);
      }
   }
   
   sum2 = 0.0;
   for (i=0; i<detdim; i++) {
      if (soft_gr[i] < 0.0) det_out[i] = 0;
      det_out[i] /= 255.0;
      sum2 += det_out[i];
   }
   loss1 /= 255.0;
/*   for (i=0; i<detdim; i++) det_out[i] /= sum2;
*/
/*   printf("pixelfold: sum_in: %10.4lf,     loss: %10.4lf,    sum_out: %10.4lf, average valsum: %6.4lf\n", 
      sum1, loss1, sum2, valsum/valnum);
*/      
   return((int)(sum2));
}

/* *********************************************************** */

double rmscalc(float shad[], float soft_gr[])
{
   int i, j;
   double m0, m1, m2, rms, var;
   
   m0 = m1 = m2 = 0.0;
   
   for (i=0; i<detdim; i++) {
      if (soft_gr[i] <= 0.0) continue;
      m0 += 1.0;
      m1 += shad[i];
      m2 += shad[i] * shad[i];
   }
   
   m1 /= m0;
   m2 /= m0;

   var = m2 - m1 * m1;
   if (var > 0.0) rms = sqrt(var);
   else rms = 0.0;
   
   return(rms);
}

/* *********************************************************** */

double rmscalc_i(short shad[], float soft_gr[])
{
   int i, j;
   double m0, m1, m2, rms, var;
   
   m0 = m1 = m2 = 0.0;
   
   for (i=0; i<detdim; i++) {
      if (soft_gr[i] <= 0.0) continue;
      m0 += 1.0;
      m1 += (double)(shad[i]);
      m2 += (double)(shad[i]) *(double)(shad[i]);
   }
   
   m1 /= m0;
   m2 /= m0;

   var = m2 - m1 * m1;
   if (var > 0.0) rms = sqrt(var);
   else rms = 0.0;
   
   return(rms);
}

/* ********************************************************** */

void clean(float sky_r[skydim], float skyim[skydim])
{
   int i, j, n, k, m, max_j[6] = {-1, -1, -1, -1, -1, -1};
   int indx[30] = { 5, 251, 259, 507, 513, 763, 767, 1019, 1021, 1275,
                    4, 250, 258, 506, 512, 762, 766, 1018, 1020, 1274,
                    6, 252, 260, 508, 514, 764, 768, 1017, 1022, 1276 };
   float b, v, max[6] = {-100000.0, -100000.0, -100000.0, -100000.0, -100000.0, -100000.0};
   
   
   for (i=0; i<skydim; i++) skyim[i] = 0.0;

   for (i=1276; i<skydim-1276; i++) {
      for (j=0; j<30; j++) {
         v = sky_r[i+indx[j]];
         for (n=0; n<6; n++) {
            if (v > max[n]) {
               for (k=5; k>n; k--) {
                  max[k] = max[k-1];
                  max_j[k] = max_j[k-1];
               }
               max[n] = v;
               max_j[n] = i+indx[j];
               goto nextv1;
            }
         }
nextv1: 
         v = sky_r[i-indx[j]];
         for (n=0; n<6; n++) {
            if (v > max[n]) {
               for (k=5; k>n; k--) {
                  max[k] = max[k-1];
                  max_j[k] = max_j[k-1];
               }
               max[n] = v;
               max_j[n] = i+indx[j];
               goto nextv2;
            }
         }
      }
nextv2: 
      b = 0.0;
      n = 0;
      for (j=0; j<30; j++) {
         k = i+indx[j];
         for (m=0; m<6; m++) if (k == max_j[m]) goto try2;
         b += sky_r[k];
         n++;
try2:    k = i-indx[j];
         for (m=0; m<6; m++) if (k == max_j[m]) continue;
         b += sky_r[k];
         n++;
      }
      
      b /= (float)(n);
      
      skyim[i] = sky_r[i] - b;
   }
   
   return;
}

/* ********************************************************** */

void clean2(float sky_r[skydim], float xx[], float yy[], int n_common, int backmodels, float skyim[skydim])
{
   int i, j, n, k, m, jj, ix, iy, ii;
   int indx[10] = { 5, 251, 259, 507, 513, 763, 767, 1019, 1021, 1275 };
   float v, sky_l[skydim], b;
   double r[25], nr[25], r2[25], rc[25], nrc[25], rc2[25];
   double dx, dy, dj, dk, rr;
   
   for (i=0; i<skydim; i++) {
      skyim[i] = 0.0;
      sky_l[i] = sky_r[i];
   }
   for (i=0; i<25; i++) r[i] = nr[i] = r2[i] = rc[i] = nrc[i] = rc2[i] = 0.0;

   for (n=0; n<n_common; n++) {
      ix = (int)(xx[backmodels+n] + 0.5);
      iy = (int)(yy[backmodels+n] + 0.5);
      ii = iy + ix * sky_ydim;
      v = 0.0;
      m = 0;
      for (i=0; i<10; i++) {
         j = ii + indx[i];
         if (j < 0) continue;
         if (j >= skydim) goto conti;
         v += sky_r[j];
         m++;
conti:   j = ii - indx[i];
         if (j < 0) continue;
         if (j >= skydim) continue;
         v += sky_r[j];
         m++;
      }
      if (m > 0) v /= (float)(m); else continue;
      
      for (i=-2; i<3; i++) {
         for (j=-2; j<3; j++) {
            jj = ii + j * sky_ydim + i;
            sky_l[jj] = v;	/* replace peak by average level */
         }
      }
   }
   

   for (ii=1275; ii<skydim-1275; ii++) {

      v = 0.0;
      m = 0;
      for (i=0; i<10; i++) {
         j = ii + indx[i];
         if (j < 0) continue;
         if (j >= skydim) goto conti2;
         v += sky_l[j];
         m++;
conti2:  j = ii - indx[i];
         if (j < 0) continue;
         if (j >= skydim) continue;
         v += sky_l[j];
         m++;
      }
      if (m > 0) v /= (float)(m); else continue;
      
      skyim[ii] = sky_l[ii] - v;	/* remove local mean level */
   }
   
   dx = xx[backmodels] - 127.0;
   dy = yy[backmodels] - 127.0;

   for (i=0; i<skydim; i++) {
      j = i / sky_ydim;
      k = i % sky_ydim;
      dj = j - 127;
      dk = k - 127;
      rr = sqrt((dx - dj) * (dx - dj) + (dy - dk) * (dy - dk)) / 6.0;
      m = rr;
      rr = sqrt(dj * dj + dk * dk);
      if (m > 24) m = 24;
      if (rr > 126.0) continue;
      r[(int)(rr/5.1)] += skyim[i];
      r2[(int)(rr/5.1)] += skyim[i] * skyim[i];
      nr[(int)(rr/5.1)] += 1.0;
      rc[m] += skyim[i];
      rc2[m] += skyim[i] * skyim[i];
      nrc[m] += 1.0;
   }
   for (i=0; i<25; i++) {
/*      printf("%5d", i); */
      if (nr[i] > 0.0) r2[i] = sqrt(r2[i] * nr[i] - r[i] * r[i]) / nr[i];
      if (nrc[i] > 0.0) rc2[i] = sqrt(rc2[i] * nrc[i] - rc[i] * rc[i]) / nrc[i];
   }

/*   printf("\n");
   printf("Mean sky image levels around source at position: %7.3lf, %7.3lf\n", dx, dy); 
   for (i=0; i<25; i++) 
      if (nr[i] > 0.0) printf("%5.1lf", r[i]/nr[i]);
      else printf("     ");
   printf("\n");
   for (i=0; i<25; i++) 
      if (nr[i] > 0.0) printf("%5.1lf", r2[i]/sqrt(nr[i]));
      else printf("     ");
   printf("\n");
   for (i=0; i<25; i++) printf("%5.0lf", nr[i]);
   printf("\n");
   
   printf("Mean sky image levels around position: 0.000  0.000\n");
   for (i=0; i<25; i++) 
      if (nrc[i] > 0.0) printf("%5.1lf", rc[i]/nrc[i]);
      else printf("     ");
   printf("\n");
   for (i=0; i<25; i++) 
      if (nrc[i] > 0.0) printf("%5.1lf", rc2[i]/sqrt(nrc[i]));
      else printf("     ");
   printf("\n");
   for (i=0; i<25; i++) printf("%5.0lf", nrc[i]);
   printf("\n");
*/
   return;
}

/* ********************************************************** */

void peaktransfer(float x, float y, float sky_r[skydim], float impeak[skydim])
{
   int ix, iy, ixy, jj;
   int i, j, n, k;
   float b, v;
   
   ix = x + 0.5;
   iy = y + 0.5;
   ixy = ix * sky_ydim + iy;	/* index for peak center */
   
   for (i=-2; i<3; i++) {
      for (j=-2; j<3; j++) {
         jj = ixy + j * sky_ydim + i;
         impeak[jj] = sky_r[jj];	/* save peak counts into 'impeak'-image */
      }
   }
}

/* ************************************************************************************* */

void fi_headr(char fitsfil[])
{
   FILE *fi_fil;
   char fi_line[81];
   int i, j;
   
   fi_fil = fopen(fitsfil, "wt");
   
   sprintf(fi_line, "SIMPLE  =                    T / Written by midi_b5 version March 11, 2004  ");    
   for (i=74; i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   fprintf(fi_fil, "%s", fi_line);
   sprintf(fi_line, "BITPIX  =                    8 /  ");    
   for (i=32; i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   fprintf(fi_fil, "%s", fi_line);
   sprintf(fi_line, "NAXIS   =                    0 /  ");    
   for (i=32; i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   fprintf(fi_fil, "%s", fi_line);
   sprintf(fi_line, "EXTEND  =                    T / File contains extension  ");    
   for (i=56; i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   fprintf(fi_fil, "%s", fi_line);
   sprintf(fi_line, "DATGROUP=              UNKNOWN /  ");    
   for (i=32; i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   fprintf(fi_fil, "%s", fi_line);
   sprintf(fi_line, "END  ");    
   for (i=3; i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   fprintf(fi_fil, "%s", fi_line);
   
   for (i=0;  i<80; i++) fi_line[i] = 32;
   fi_line[80] = 0;
   for (j=6; j<36; j++) fprintf(fi_fil, "%s", fi_line);

   fclose(fi_fil);
}

/* **************************************************************************************** */

