 /*
Reference shadowgram preparation. Niels Lund Jan. 2003

!!!! 2003-04-28 Updated for use on Unix and away from development environment !!!!
!!!! Adapt to your local environment by providing correct "sweethome"-string  !!!!

May 9 2003: modified with new energy bands, emphazising the low energies

Uses pre-calculated shadowgrams "shadgram" in four energy bands. (prepared by ipfspec)  [yymmdd_#shdJ.bin]

The statementss with *** describes my plan. It did not work! For the moment only the pixelfold, smooth and
sensi routines are used to generate the reference shadowgrams. These therefore contains both internal and
sky diffuse background.

***Uses the auxiliary data prepared by "skymapi" containing a diffuse background shadowgram.
***This shadowgram is smoothed according to the position determination uncertainty of the microstrip
***detector to generate more realistic diffuse sky shadowgrams in four energy bands.
***The best fitting diffuse sky shadowgrams are then subtracted from the reference shadowgram and the residual 
***(in four energy bands) is taken as a basis for models for the internal background shadowgrams.

The internal background shadowgrams are smoothed to remove faint sources and reduce the statistical noise
and finally regions with excessive background is flagged as dont-use regions for the later analyses. 
The resulting shadowgrams are output in the "reference1/2.bin" files.

*/


#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <string.h>
#include <time.h>

#define detdim		65536		/* the detector is divided into 256*256 pixels */
#define sky_xdim	255
#define sky_ydim	255
#define sky_dim		sky_xdim*sky_ydim
#define det_depth	-55.0
#define det_xdim	256
#define det_ydim	256
#define mask_height	3399.0
#define sourcmax	2
#define SWAP(a,b) {double temp=(a);(a)=(b);(b)=temp;}	/* function used in "gaussj_nr" */

/*	
*	NOTE: The jemx variable is 0 for JEM-X1 and 1 for JEM-X2 in these routines !!
*/

/* int backpro(int ee, int *soft_map); */
int sh_update(unsigned short x, unsigned short y, int jemx);
int sh_xy(double x_rms, double y_rms, float scatt_list[], short scatt_delta[]);
int sign(int x);
int shapefold(short det[], float comb_list[], short comb_delta[], int comb_len, float det_out[]);
int gaussj_nr(double a[], int n, double b[], int m);
float fun(int k, int ii, int ee, int active);
double imagefit(float shdg[], int jemx, int ee, float resid[], int n_source);
int pixelfold(short det[], int pixel_st[], unsigned char pixel_list[], float det_out[]);
void smooth(int jemx, double accumT);
int sensi(int ee, int num, float ss[][]);


/* program to generate JEM-X backgrounnd models based on input from the "skymap" 
   program and from flight data taken on empty fields.

*/

unsigned d_allow_st[detdim];
/*	The d_allow_st array gives for each 16-bit pixel in the 
   	detector the 32-bit address of the corresponding list 
        in the d_allow array.
	(generated by the skymap routine)
*/

short d_allow_l[detdim];
/*	array giving the length of the list in the d_allow array
   	corresponding to a given detector pixel
        */
int d_allow_p[detdim];
/*	array giving the address of the first sky pixel visible 
   	from a given detector pixel 
*/

short initmap[detdim]; 
/*	array with "ones" in all useable pixel positions and "minus one" in all positions 
        outside the active area of the microstrip plate. Other "dont use" regions are 
        marked with negative numbers. (hot pixels near calibration sources: -2,
        dead anodes: -3, unstable anodes: -4, areas with excessive background with -5).
*/

short ref_shdgr[4][detdim];
/*	Reference shadowgrams from JEM-X observations of an empty field.
   	These shadowgrams are integrated over 4 energy bands: 
        3-11, 11-22, 22-35, 35-50 keV.
*/

double ref_tot[25];
/*   	Parameters characterizing the "ref_shdgr" arrays:
   	ref_tot[0]: total number of events in the ref_shdr[0][] array
   	ref_tot[1]: total number of events in the ref_shdr[1][] array
   	ref_tot[2]: total number of events in the ref_shdr[2][] array
   	ref_tot[3]: total number of events in the ref_shdr[3][] array
   	ref_tot[23]: Start integration time (days since 2002:00:0000) for the ref_shdr[][] arrays
   	ref_tot[24]: Stop integration time (days since 2002:00:0000) for the ref_shdr[][] arrays
   	ref_tot[17-21]: pulseheight boundaries for the four energy ranges
           		 used to select the four energy ranges for the arrays
*/

float diff_shdgr[4][detdim];
/*	arrays containing the shadowgrams calculated for a uniform diffuse 
        background. 
        Each shadowgram is normalized to contain a total of 1.0 event.
*/

float soft_shdgr[4][detdim];
/*	Smoothed shadowgrams derived by this program by pixelfolding 
   	the reference empty field observations.
   	These shadowgrams corresponds to the 4 energy bands: 
        3-11, 11-22, 22-35, 35-50 keV.
        "Dont-use" regions have been defined around the four calibration 
        sources in the smoothed shadowgrams, and for other pixels with excessive
        noise counts. Optionally the dead and unstable anodes are also indicated 
        by negative values:
        -1: Calibration source
        -2: Dead anodes (and low efficiency neighbors)
        -3: Unstable anodes
        -4: Excessive noise
        The smoothing is done by averaging over a 3x3 set of 
        pixels (selected to be oriented approximately perpendicular to 
        the direction to the center of the detector at any pixel. This 
        is intended to preserve the radid radial variations near the 
        edge of the detector at the expense of any azimuthal variations.)
        Each shadowgram is normalized to contain a total of 1.0 event.
*/

double soft_tot[30];
/*   	Parameters characterizing the "soft_shdgr" arrays:
   	soft_tot[0]: total number of events behind the soft_shdr[0][] array
   	soft_tot[1]: total number of events behind the soft_shdr[1][] array
   	soft_tot[2]: total number of events behind the soft_shdr[2][] array
   	soft_tot[3]: total number of events behind the soft_shdr[3][] array
   	soft_tot[4]: Integration time (s) behind the soft_shdr[][] arrays
   	soft_tot[5-7]:   x, y and r for excluded region around cal. src. 1
   	soft_tot[8-10]:  x, y and r for excluded region around cal. src. 2
   	soft_tot[11-13]: x, y and r for excluded region around cal. src. 3
   	soft_tot[14-16]: x, y and r for excluded region around cal. src. 4
   	soft_tot[17-21]: pulseheight boundaries for the four energy ranges
           		 used to select the four energy ranges for the arrays
        soft_tot[22] = fac0: number of pixels used in smoothing
        soft_tot[23] = ref_tot[23]: start time of reference data (days since 2002:01:01
        soft_tot[24] = ref_tot[24]; stop time of reference data (days since 2002:01:01
   	soft_tot[25]:  1: dead anodes included in soft_shdgr, 0: dead anodes excluded
        soft_tot[26-29]: number of remaining useful pixels in soft_shdgr
*/

double parameters[20];	/* array with the specific  parameters for current skymap:
        parameters[0] = byt_index;	length of 'd_allow' list 
        parameters[1] = lsum;		sum of ('det_pixel'*'sky pixels visible from the det_pixel') 
        parameters[2] = isum;		number of active 'det_pixels'
        parameters[3] = skyxdim;	number of X-det_pixels in sky region of interest 
        parameters[4] = skyxoffs; 	X-displacement of corner of sky region of interest from on-axis
        parameters[5] = skyydim;	number of Y-det_pixels in sky region of interest 
        parameters[6] = skyyoffs; 	Y-displacement of corner of sky region of interest from on-axis
        parameters[7] = cond;		condition parameter for inclusion of collimator in backprojection 
	          		cond = 0: disregard collimator in backprojection 
                                cond = 1: include collimator
        parameters[8] = jemx; 		JEM-X unit number 
        parameters[9] = FOV_radius	Radius of FOV expressed in sky-pixels (the unit used for skyxdim & skyydim)
        parameters[10] = indep;		interaction depth of photons in the detector 
           			Common-X is the direction from JEM-X2 towards JEM-X1 
        parameters[11] = alfa; 		angle between mask X-axis and Common-X 
        parameters[12] = beta;		angle between collimator X-axis and Common-X  		
        parameters[13] = gamma;		angle between microstrip X-axis and Common-X 

        parameters[14] = 0.0;		was 'Hsource' for ESTEC calibrations (not used)
        parameters[15] = xshi;		X-shift of microstrip center relative to collimator center 
        parameters[16] = yshi;		Y-shift of microstrip center relative to collimator center 
        parameters[17] = maskoffs;	displacement of mask center spacecraft centerline, now not used
        parameters[18] = mask_sep;	height of mask above detector window 
        parameters[19] = 0.0;		not used 
*/

short maskhole[22501];
char notes[1000];	/* notes & comments for current skymap */
char sweethome[256];	/* home directory string */

int pixel_st[detdim];
unsigned char pixel_list[500000];


double calsrcx[2][4] = {38.25, 103.25, 152.25, 216.75, 38.25, 103.25, 152.25, 217.75}; 
double calsrcy[2][4] = {197.5, 235.0, 234.25, 197.0, 197.0, 235.0, 234.5, 197.0}; 
double calrad[2][4] = {7.5, 6.0, 6.5, 6.0, 6.0, 6.0, 6.0, 6.0};
float b_amppos[21] =    {0.,15.,29.,43.,55.,67.,79.,91.,103.,115.,127.,139.,151.,163.,175.,187.,199.,211.,225.,239.,255.};

int pha_sel[5];

int a_reject = 1;		/* default: reject dead anodes */

int iorder, yshift = 0, m_max;
int nr, nx, ny;
unsigned  byt_index, allowmax, sky_index;
unsigned loss[20], distri[100], gain[20];
double alfarot, betarot, gammarot;
double solut[3], x_shi, y_shi, snratio[4][101];	
float fun_shdgr[detdim], refshad[detdim], limit[4][101];

FILE *ofil, *ifil, *illumfil, *gnuxfil, *gnuyfil;

/* ************************************************************* */

int main(int argc, char* argv[])
{
/*	
*	NOTE: The jemx variable is 0 for JEM-X1 and 1 for JEM-X2 in these routines !!
*/
        char skyfil[200], skofil[200], ref_fil[200], pixfil[200], deadfil[200], iline[300];
        char signs[40] = { '0','1','2','3','4','5','6','7','8','9',
           		   'a','b','c','d','e','f','g','h','i','j',
                           'k','l','m','n','o','p','q','r','s','t',
                           'u','v','x','y','z','+','+','+','+','+' };
        
        short scatt_delta[200];
	int scatt_len, deadanod[100], deadanod_type[100], deadanod_num = 0;
        unsigned short iy, jx;

   	int i, j, ii, k, normax, normin, bn, il_col[49], irsky, jemx, softsum, det_index;
        int normtal[20], kk, bi, normimin[4], normimax[4], refsum, pixindex;
        int byt_index, lsum, isum, skyxdim, skyxoffs, skyydim, skyyoffs, cond, n_source;
        int b_org1[256], b_org2[256], b_org3[256], c_org2[256];
   	int m, n, det_x0, det_y0, det_xy0, limlow, limhigh, jval;
        int ix, iix, iiy, isx, isy, ipx, ipy, pjemx;
        int coll_cnts[7][7][7][7];

        float back1[256], back2[256], back3[256], backm1[256], backm2[256], backm3[256];
        float cath2[256], cathm2[256];
        float b_fun2[256], c_fun2[256];
        float b_res2[256], c_res2[256];
        float bc_val[256], bc0, bc_fact;
        float scatt_list[200];
        float coll_illum[7][7][7][7], coll_ideal[7][7][7][7], c_ideal[7][7], c_illum[7][7];
        float min_il[7][7], min_id[7][7], nonzer[7][7], max_il[7][7], max_id[7][7];
        float glob_id[7][7], glob_il[7][7];
        float off_il, off_id, scl_il, scl_id, cx_cnts[7], cy_cnts[7];
        
        double abs_len, escape, xy_scatt[21][21], x_rms, y_rms, m0, m1, m2, accumT; 
        double sum_illum[7][7], sum_ideal[7][7];
        double sum, rms, m1ref, m2ref, m1pix, m2pix, m1soft, m2soft, rmsref, rmspix, rmssoft;
        double normdist[20], normind[4][20], colnorm, xxdet, yydet, rrdet, rdet2;
        double indep, FOV_radius, alfa, beta, gamma, maskoffs, mask_sep, softsm[4];

/* !!!!!!!!!!!!!!!!!!!!!!!! provide your own home directory for "sweethome" !!!!!!!!!!!!!!!!!! */         

	sprintf(sweethome, "%s", "/r2/jemx/nlwork"); 
/*	sprintf(sweethome, "%s", "/home/nl/jemx/ipfread/jemx1-2"); */

/* !!!!!!!!!!!!!!!!!!!!!!!! provide your own home directory for "sweethome" !!!!!!!!!!!!!!!!!! */         

        rdet2 = 124.5 * 124.5;
        
        jemx = 2;		/* take JEM-X2 as default */

        for (i=0; i<21; i++) b_amppos[i] *= 0.975;
        for (i=0; i<21; i++) b_amppos[i] += 5.0;

        for (i=0; i<7; i++)
           for (j=0; j<7; j++) {
              c_ideal[i][j] = c_illum[i][j] = 0.0;
              for (k=0; k<7; k++)
                 for (m=0; m<7; m++) 
                    coll_cnts[i][j][k][m] = coll_illum[i][j][k][m] = coll_ideal[i][j][k][m] = 0.0;
           }

        for (j=0; j<4; j++)
           for (i=0; i<detdim; i++) diff_shdgr[j][i] = soft_shdgr[j][i] = 0.0;

        x_rms = 0.01;
        y_rms = 0.01;
        bc_fact = 0.0;	/* back-amplifier modulation function amplitude */

        if (argc > 1) jemx = atoi(argv[1]);
        else {printf("parameters: jemx shadowgram_file[.bin]\n"); return;}

        sprintf(ref_fil, "%s/reference/period5new/021225_0_0050_shd%1d.bin", sweethome, jemx+1);
        sprintf(skyfil, "%s/reference/skmapic%1d_0", sweethome, jemx+1);

        if (argc > 2) sprintf(ref_fil, "%s", argv[2]);

        sprintf(pixfil, "%s/reference/pixdata%1d.bin", sweethome, jemx+1);
        ifil = fopen(pixfil, "rb");
        if (ifil == NULL) {printf(" Error in opening %s\n", pixfil); return;}
        fread(pixel_st, 4, detdim, ifil);
        pixindex = pixel_st[0];
        fread(pixel_list, 1, pixindex, ifil);
        fclose(ifil);

        printf("Reading reference (empty field) maps from %s.\n", ref_fil);
        ifil = fopen(ref_fil, "rb");
        if (ifil == NULL) {printf(" Error in opening %s\n", ref_fil); return;}
        fread(ref_shdgr, 2, 4*detdim, ifil);	/* get four reference background shadowgrams */
        fread(ref_tot, 8, 25, ifil);
        fclose(ifil);

        pha_sel[0] = ref_tot[17];
        pha_sel[1] = ref_tot[18];
        pha_sel[2] = ref_tot[19];
        pha_sel[3] = ref_tot[20];
        pha_sel[4] = ref_tot[21];
        printf("pha_sel as read from bin file: %5d %5d %5d %5d %5d\n",
           pha_sel[0], pha_sel[1], pha_sel[2], pha_sel[3], pha_sel[4]);
           
        k = 0;
        printf("Refsums: ");
        for (j=0; j<4; j++) {
           refsum = 0;
           for (i=0; i<detdim; i++) refsum += ref_shdgr[j][i];
           printf(" %8d,  file: %8.1lf", refsum, ref_tot[j]);
           if (refsum > 0) k++;
        }
        accumT = (double)(ref_tot[24]-ref_tot[23])*86400.0;
        printf("\nIntegration interval from: %7.3lf to %7.3lf,  Integration time: %8.1lf\n", 
           ref_tot[23], ref_tot[24], accumT);
         
        if (k == 0) return;

        strcpy(skofil, skyfil);
        strcat(skofil, ".aux");
        printf("Reading sky model data from %s.  ", skofil);
        ofil = fopen(skofil, "rb");
        if (ofil == NULL) {printf(" Error in opening %s\n", skofil); return;}
        fread(parameters, 8, 20, ofil);
        fread(d_allow_st, 4, detdim, ofil);
        fread(d_allow_l, 2, detdim, ofil);
        fread(d_allow_p, 4, detdim, ofil);
        fread(maskhole, 2, 22501, ofil);
        fread(notes, 1, 1000, ofil);
        fclose(ofil);

        byt_index = parameters[0];	/* length of 'd_allow' list */
        lsum = parameters[1]; 		/* sum of ('det_pixel'*'sky pixels visible from the det_pixel') */
        isum = parameters[2]; 		/* number of active 'det_pixels' */
        skyxdim = parameters[3]; 	/* number of X-det_pixels in sky region of interest */
        skyxoffs = parameters[4]; 	/* X-displacement of corner of sky region of interest from on-axis */
        skyydim = parameters[5]; 	/* number of Y-det_pixels in sky region of interest */
        skyyoffs = parameters[6]; 	/* Y-displacement of corner of sky region of interest from on-axis */
        cond = parameters[7]; 		/* condition parameter for inclusion of collimator in backprojection */
        pjemx = parameters[8]; 		/* JEM-X unit number */
        FOV_radius = parameters[9];	/* Radius of FOV expressed in sky-pixels (the unit used for skyxdim & skyydim) */
        indep = parameters[10]; 	/* interaction depth of photons in the detector */
           		/* Common-X is the direction from JEM-X2 towards JEM-X1 */
        alfa = parameters[11]; /*  		angle between mask X-axis and Common-X */
        beta = parameters[12]; /* 		angle between collimator X-axis and Common-X */ 		
        gamma = parameters[13]; /* 		angle between microstrip X-axis and Common-X */
        i = parameters[14]; 		/* not used */
        x_shi = parameters[15]; 		/* X-shift of microstrip center relative to collimator center */
        y_shi = parameters[16]; 		/* Y-shift of microstrip center relative to collimator center */
        maskoffs = parameters[17]; 	/* displacement of mask center spacecraft centerline, now not used */
        mask_sep = parameters[18]; 	/* height of mask above detector window */
        i = parameters[19];  		/* not used */
        
        printf("jemx %1d, calrad: %6.2lf %6.2lf %6.2lf %6.2lf, x/y_shift: %+4.1lf/%+4.1lf\n", 
           jemx+1, calrad[jemx][0], calrad[jemx][1], calrad[jemx][2], calrad[jemx][3], x_shi, y_shi);

        for (iy=0; iy<256; iy++) {
           for (jx=0; jx<256; jx++) {
              det_index = jx * 256 + iy;
              if (sh_update(jx, iy, jemx) > 0) initmap[det_index] = 1;
              else initmap[det_index] = -2;
              yydet = (double)(iy) - 127.0 + y_shi;
              xxdet = (double)(jx) - 127.0 + x_shi;
              rrdet = xxdet * xxdet + yydet * yydet;
              if (rrdet > rdet2) initmap[det_index] = -1;
           }
        }
        
        if (pjemx != jemx) {
           printf("Warning: jemx unit number from %s does not match specied jemx. Continue? (y/n): ", skofil);
           if (getchar() != (int)('y')) {printf("\n"); return;}
           printf("OK, continuing!\n");
        }
        
        soft_tot[25] = 1.0 - a_reject;
/*        printf("Eliminate dead anodes? (y/n): ");
        if (getchar() != (int)('y')) {printf("\n"); goto skipdead;}
        printf("\n"); 
*/
        soft_tot[25] = 0.0;		/* indicate elimination of dead anodes */
        sprintf(skofil, "%s/reference/dead_anodes%1d.txt", sweethome, jemx+1);
	printf("%s\n", skofil);
        ofil = fopen(skofil, "rt");
        if (ofil == NULL) {printf(" Error in opening %s\n", skofil); return;}
        fgets(iline, 100, ofil);
	printf("read line %3d:  %s", -1, iline);

        sscanf(iline, "%d %d ", &deadanod_num, &j);

        printf("Read information about %3d dead anodes from file: %s\n", deadanod_num, skofil);

        for (i=0; i<deadanod_num; i++) {
           fgets(iline, 100, ofil);
	   printf("read line %3d:  %s", i, iline);

           sscanf(iline, "%d %d ", &deadanod[i], &deadanod_type[i]);
           printf("%2d anode#: %3d,  cause: %1d   %s\n", i, deadanod[i], deadanod_type[i], iline);

           for (iy=0; iy<256; iy++) {
                 det_index = deadanod[i] * 256 + iy;
                 if (initmap[det_index] < 0) continue;
                 if (deadanod_type[iy] == 4) initmap[det_index] = -4; else initmap[det_index] = -3;
           }

        }
        fclose(ofil);
        printf("Read information about %3d dead anodes from file: %s\n", deadanod_num, skofil);

skipdead:        
/*        gnuyfil = fopen("ydistr.gnu", "wt");
        gnuxfil = fopen("xdistr.gnu", "wt");
        illumfil = fopen("coll_illum.txt", "wt");
*/
        jval = 0;
        for (isx=0; isx<7; isx++) cx_cnts[isx] = cy_cnts[isx] = 0;
nextjv:
   	pixelfold(&ref_shdgr[jval][0], pixel_st, pixel_list, refshad);
        n_source = 2;
        for (k=0; k<1; k++) {
              for (n=0; n<1; n++) {
                      
              scatt_len = sh_xy(x_rms, y_rms, scatt_list, scatt_delta);
              shapefold(d_allow_l, scatt_list, scatt_delta, scatt_len, fun_shdgr);
              sum = 0.0;
              for (i=0; i<detdim; i++) if (initmap[i] > 0) sum += fun_shdgr[i];
              for (i=0; i<detdim; i++) {
                 fun_shdgr[i] /= sum;
                 ix = i / 256;
                 iy = i % 256;
                 iix = ix - 2 + 35;	/* place center of 35x35 pixel subsquare at X = 128 */
                 iiy = iy - 5 + 35;	/* place center of 35x35 pixel subsquare at Y = 124 (y_shi offset) */
                 isx = iix / 35;	/* subsquare index */
                 isy = iiy / 35;
                 ipx = iix % 7;	/* pixel index inside collimator channel */
                 ipy = iiy % 7;	/* pixel index inside collimator channel */
                 isx -= 1;
                 isy -= 1;
                 if ((isx>=0) && (isx<7) && (isy>=0) && (isy<7) && (initmap[i]>0)) {
                    coll_illum[isx][isy][ipx][ipy] += refshad[i];
                    coll_cnts[isx][isy][ipx][ipy] += 1;
                    coll_ideal[isx][isy][ipx][ipy] += fun_shdgr[i];
                    if ((ix>78) && (ix<177) && (iy>78) && (iy<177)) {
                       c_ideal[(iix%7)][(iiy%7)] += fun_shdgr[i];
                       c_illum[(iix%7)][(iiy%7)] += refshad[i];
                       cx_cnts[(iix%7)]++;
                       cy_cnts[(iiy%7)]++;
                    }
                 }
              }
                 
/*              rms = imagefit(refshad, jemx, jval, &soft_shdgr[jval][0], n_source);
*/
              for (i=0; i<detdim; i++) 
                 if (initmap[i] > 0) soft_shdgr[jval][i] = refshad[i];
                 else soft_shdgr[jval][i] = initmap[j];
              softsm[jval] = 0.0;
              for (i=0; i<detdim; i++) 
                 if (initmap[i] > 0) softsm[jval] += soft_shdgr[jval][i];              
  printf("x_rms: %4.2lf, y_rms: %4.2lf, strength: %8.2lf, residual shadowgram rms: %8.3lf,  shdgr. cnts: %8.3lf\n", 
                 x_rms, y_rms, solut[1], rms, softsm[jval]);
           }
        }

        m0 = m1ref = m2ref = m1pix = m2pix = m1soft = m2soft = 0.0;
        for (iy=0; iy<256; iy++) {
           for (jx=0; jx<256; jx++) {
              det_index = jx * 256 + iy;
              yydet = (double)(iy) -127.0 + y_shi;
              xxdet = (double)(jx) -127.0 + x_shi;
              rrdet = xxdet * xxdet + yydet * yydet;
              if (rrdet < rdet2) {
                 m0 += 1.0;
                 m1ref += ref_shdgr[jval][det_index];
                 m2ref += ref_shdgr[jval][det_index] * ref_shdgr[jval][det_index];
                 m1pix += refshad[det_index];
                 m2pix += refshad[det_index] * refshad[det_index];
                 m1soft += soft_shdgr[jval][det_index];
                 m2soft += soft_shdgr[jval][det_index] * soft_shdgr[jval][det_index];
              }
           }
        }
        
        m1ref /= m0;
        m2ref /= m0;
        m1pix /= m0;
        m2pix /= m0;
        m1soft /= m0;
        m2soft /= m0;
        
        rmsref = sqrt(m2ref - m1ref*m1ref);
        rmspix = sqrt(m2pix - m1pix*m1pix);
        rmssoft = sqrt(m2soft - m1soft*m1soft);
        
        printf("RMS values: ref: %6.3lf,  pix: %6.3lf,  soft: %6.3lf\n", rmsref, rmspix, rmssoft);
        
        limlow = calsrcx[jemx][0] + calrad[jemx][0] + 0.5;
        limhigh = calsrcx[jemx][1] - calrad[jemx][1] - 0.5;
        for (iy=0; iy<256; iy++) {
           back1[iy] = backm1[iy] = b_org1[iy] = 0;
           for (jx=0; jx<256; jx++) {
              if ((jx >= limlow) && (jx <= limhigh)) {
                 det_index = jx * 256 + iy;
                 back1[iy] += refshad[det_index];
                 b_org1[iy] += ref_shdgr[jval][det_index];
              }
           }  
        }

        limlow = calsrcx[jemx][1] + calrad[jemx][1] + 0.5;
        limhigh = calsrcx[jemx][2] - calrad[jemx][2] - 0.5;
        for (iy=0; iy<256; iy++) {
           back2[iy] = backm2[iy] = b_org2[iy] = 0;
           for (jx=0; jx<256; jx++) {
              if ((jx >= limlow) && (jx <= limhigh)) {
                 det_index = jx * 256 + iy;
                 back2[iy] += refshad[det_index];
                 backm2[iy] += d_allow_l[det_index];
                 b_org2[iy] += ref_shdgr[jval][det_index];
                 b_fun2[iy] += fun_shdgr[det_index];
                 b_res2[iy] += soft_shdgr[jval][det_index];
              }
           }  
        }
        for (jx=0; jx<256; jx++) {
           cath2[jx] = cathm2[jx] = c_org2[jx] = 0;
           for (iy=0; iy<256; iy++) {
              if ((iy >= limlow) && (iy <= limhigh)) {
                 det_index = jx * 256 + iy;
                 cath2[jx] += refshad[det_index];
                 cathm2[jx] += d_allow_l[det_index];
                 c_org2[jx] += ref_shdgr[jval][det_index];
                 c_fun2[jx] += fun_shdgr[det_index];
                 c_res2[jx] += soft_shdgr[jval][det_index];
              }
           }  
        }

        limlow = calsrcx[jemx][2] + calrad[jemx][2] + 0.5;
        limhigh = calsrcx[jemx][3] - calrad[jemx][3] - 0.5;
        for (iy=0; iy<256; iy++) {
           back3[iy] = backm3[iy] = b_org3[iy] = 0;
           for (jx=0; jx<256; jx++) {
              if ((jx >= limlow) && (jx <= limhigh)) {
                 det_index = jx * 256 + iy;
                 back3[iy] += refshad[det_index];
                 b_org3[iy] += ref_shdgr[jval][det_index];
              }
           }  
        }
        
        bc0 = 1.0 - bc_fact;
        bc_fact *= 4.0;

        for (iy=0; iy<256; iy++) {	/* generate modulation according to position relative to  back-amplifiers */
           for (j=0; j<21; j++) if (b_amppos[j] >= iy) goto found;
found:	   if (j == 0) bc_val[iy] = 1.0 - bc_fact / 4.0;
   	   else {
              if ((float)(iy) > ((float)(b_amppos[j-1]+b_amppos[j])/2.0)) 
                 bc_val[iy] = bc0 - bc_fact * (float)(b_amppos[j]-iy)/(float)(b_amppos[j-1]-b_amppos[j]);
              else
                 bc_val[iy] = bc0 - bc_fact * (float)(iy-b_amppos[j-1])/(float)(b_amppos[j-1]-b_amppos[j]);
           }
        }

        m0 = m1 = m2 = 0.0;
        k = 0;
        if (jval == 3) k = 400;
        for (iy=0; iy<256; iy++) {
           if (iy < 245) {
              m0++;
              m1 += back2[iy]*bc_val[iy];
              m2 += back2[iy] * bc_val[iy] * back2[iy] * bc_val[iy];        
           }
/*           fprintf(gnuyfil, "%3d  %8.3f\n", iy, (float)(back2[iy]*bc_val[iy])+k); */ /* 0, 10, 12, 14 pixelfold */

        }
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %5d\n", iy, b_org2[iy]+k); */	/* 1, 11, 13, 15 original */
/*        fprintf(gnuyfil, "\n\n");
        if (jval > 0) goto ysk1;
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %8.3f\n", iy, backm2[iy]); */	/* 2 d_allow_l */
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %8.3f\n", iy, b_fun2[iy]*1000.0); */	/* 3 res_fold */
/*        fprintf(gnuyfil, "\n\n");
        fprintf(gnuyfil, "%6.2f  %3d\n", (float)(0.0), 0);
        for (iy=1; iy<36; iy++) fprintf(gnuyfil, "%6.2f  %3d\n",(float)((iy*7.05)-y_shi),2000); */ /* 4 collimator */
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %8.3f\n", iy, b_res2[iy]+200); */	/* 5 */
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %8.3f\n", iy, bc_val[iy]*300.0); */	/* 6 */
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<21; iy++)  fprintf(gnuyfil, "%5.1f  %8.3f\n", b_amppos[iy], (float)(600.0)); */ /* 7 */
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %8.3f\n", iy, back1[iy]*bc_val[iy]); */	/* 8 */
/*        fprintf(gnuyfil, "\n\n");
        for (iy=0; iy<256; iy++) fprintf(gnuyfil, "%3d  %8.3f\n", iy, back3[iy]*bc_val[iy]); */	/* 9 */
/*        fprintf(gnuyfil, "\n\n");
*/
ysk1:

        m1 /= m0;
        m2 /= m0;
        rms = sqrt(m2 - m1*m1);
        printf ("back2: # of channels: %6.0lf, mean: %6.2lf, sigma: %5.2lf,  integr. time: %8.1lf s\n", 
           m0, m1, rms, accumT);

        m0 = m1 = m2 = 0.0;
        for (jx=0; jx<256; jx++) {
           if ((jx > 7) && (jx < 249)) {
              m0++;
              m1 += cath2[jx];
              m2 += back2[jx] * cath2[jx];        
           }
/*           fprintf(gnuxfil, "%3d  %8.3f\n", jx, cath2[jx]+k); */			/* 0, 5, 7, 9 pixelfold */
        }
/*        fprintf(gnuxfil, "\n\n");
        for (jx=0; jx<256; jx++) fprintf(gnuxfil, "%3d  %5d\n", jx, c_org2[jx]+k); */	/* 1, 6, 8, 10 original */
/*        fprintf(gnuxfil, "\n\n");
        if (jval > 0) goto xsk1;
        for (jx=0; jx<256; jx++) fprintf(gnuxfil, "%3d  %8.3f\n", jx, cathm2[jx]); */	/* 2 d_allow_l */
/*        fprintf(gnuxfil, "\n\n");
        for (jx=0; jx<256; jx++) fprintf(gnuxfil, "%3d  %8.3f\n", jx, c_fun2[jx]*1000.0); */	/* 3 res_fold */
/*        fprintf(gnuxfil, "\n\n");
        fprintf(gnuxfil, "%6.2f  %3d\n", (float)(0.0), 0);
        for (jx=1; jx<36; jx++)  fprintf(gnuxfil, "%6.2f  %3d\n", (float)((jx*7.05)), 2000); */	/* 4 collimator */
/*        fprintf(gnuxfil, "\n\n");
*/
xsk1:
        m1 /= m0;
        m2 /= m0;
        rms = sqrt(m2 - m1*m1);
        printf ("cath2: # of channels: %6.0lf, mean: %6.2lf, sigma: %5.2lf\n", m0, m1, rms);

        ofil = fopen("initmap.txt", "wt");
        for (iy=256; iy>185; iy--) {
           for (jx=0; jx<256; jx++) {
              det_index = jx * 256 + iy - 1;
              if (initmap[det_index] == -1) fprintf(ofil, " "); 
              else fprintf(ofil, "%1d", jemx+1);
           }
           fprintf(ofil, "\n");
        }
        fclose(ofil);
        
        for (isy=6; isy>=0; isy--) {
           for (isx=0; isx<7; isx++) {
              glob_id[isx][isy] = glob_il[isx][isy] = 0.0;
              sum_illum[isx][isy] = sum_ideal[isx][isy] = max_id[isx][isy] = max_il[isx][isy] = 0.0;
              min_id[isx][isy] = min_il[isx][isy] = 1000000.0;
              nonzer[isx][isy] = 0;
              for (ipy=6; ipy>=0; ipy--) {
                 for (ipx=0; ipx<7; ipx++) {
                    if (coll_cnts[isx][isy][ipx][ipy] > 0) {
                       coll_illum[isx][isy][ipx][ipy] /= coll_cnts[isx][isy][ipx][ipy];
                       coll_ideal[isx][isy][ipx][ipy] /= coll_cnts[isx][isy][ipx][ipy]; 
                       if (coll_illum[isx][isy][ipx][ipy] > max_il[isx][isy]) 
                          	max_il[isx][isy] = coll_illum[isx][isy][ipx][ipy];
                       if (coll_ideal[isx][isy][ipx][ipy] > max_id[isx][isy]) 
                          	max_id[isx][isy] = coll_ideal[isx][isy][ipx][ipy];
                       if (coll_illum[isx][isy][ipx][ipy] < min_il[isx][isy]) 
                          	min_il[isx][isy] = coll_illum[isx][isy][ipx][ipy];
                       if (coll_ideal[isx][isy][ipx][ipy] < min_id[isx][isy]) 
                          	min_id[isx][isy] = coll_ideal[isx][isy][ipx][ipy];
                       nonzer[isx][isy]++;
                    } else {
                       coll_illum[isx][isy][ipx][ipy] = 0.0;
                       coll_ideal[isx][isy][ipx][ipy] = 0.0;
                    }
                    sum_illum[isx][isy] += coll_illum[isx][isy][ipx][ipy];
                    sum_ideal[isx][isy] += coll_ideal[isx][isy][ipx][ipy];
                 }
              }
              
           }
        }    
          
        i = 0;
        scl_il = scl_id = 0.0;
        for (isy=6; isy>=0; isy--) {
           for (isx=0; isx<7; isx++) {
              if ((sum_illum[isx][isy] == 0.0) || 
                  (sum_illum[isx][isy] == 0.0) || 
                  (nonzer[isx][isy] == 0.0)) continue;

              scl_il += (max_il[isx][isy] - min_il[isx][isy]) / 100.0;
              scl_id += (max_id[isx][isy] - min_id[isx][isy]) / 100.0;
              i++;
           }
        }
        scl_il /= (float)(i);
        scl_id /= (float)(i);

        sum_illum[0][0] = sum_ideal[0][0] = 0.0;
   	for (ipy=6; ipy>=0; ipy--) {
           sum_illum[1][ipy] = sum_ideal[1][ipy] = 0.0;
           sum_illum[2][ipy] = sum_ideal[2][ipy] = 0.0;
        }
        for (isy=6; isy>=0; isy--) {
           for (isx=0; isx<7; isx++) {
              c_illum[isx][isy] /= cx_cnts[isx]*cy_cnts[isy];
              c_ideal[isx][isy] /= cx_cnts[isx]*cy_cnts[isy];
              sum_illum[0][0] += c_illum[isx][isy];
              sum_ideal[0][0] += c_ideal[isx][isy];
              sum_illum[1][isx] += c_illum[isx][isy];
              sum_ideal[1][isx] += c_ideal[isx][isy];
              sum_illum[2][isy] += c_illum[isx][isy];
              sum_ideal[2][isy] += c_ideal[isx][isy];
              
           }
        }    
        sum_illum[0][0] /= (100.0 * 144.0);
        sum_ideal[0][0] /= (100.0 * 144.0);
          
        if (jval > 0) goto skprint;
/*        fprintf(illumfil, 
           "Central 98 x 98 mm2 modulo 7, repeated 4 times.   Model   JEM-X%1d\n", jemx+1);
   	for (ipy=7; ipy>=0; ipy--) {
           for (ipx=0; ipx<8; ipx++) fprintf(illumfil, " %6.0f", c_ideal[(ipx%7)][(ipy%7)]/sum_ideal[0][0]);
           fprintf(illumfil, "   %7.0f   %5.0f\n", sum_ideal[2][(ipy%7)]/sum_ideal[0][0], cy_cnts[(ipy%7)]);
        }
        for (ipx=0; ipx<8; ipx++) fprintf(illumfil, " %6.0f", sum_ideal[1][(ipx%7)]/sum_ideal[0][0]);
        fprintf(illumfil, "\n");
        for (ipx=0; ipx<8; ipx++) fprintf(illumfil, " %6.0f", cx_cnts[(ipx%7)]);
        fprintf(illumfil, "\n\n");
*/
skprint:
/*        fprintf(illumfil, 
           "Central 98 x 98 mm2 modulo 7, repeated 4 times.   Observed   JEM-X%1d, E=%1d\n", jemx+1, jval);
   	for (ipy=7; ipy>=0; ipy--) {
           for (ipx=0; ipx<8; ipx++) fprintf(illumfil, " %6.0f", c_illum[(ipx%7)][(ipy%7)]/sum_illum[0][0]);
           fprintf(illumfil, "   %7.0f   %5.0f\n", sum_illum[2][(ipy%7)]/sum_illum[0][0], cy_cnts[(ipy%7)]);
        }
        for (ipx=0; ipx<8; ipx++) fprintf(illumfil, " %6.0f", sum_illum[1][(ipx%7)]/sum_illum[0][0]);
        fprintf(illumfil, "\n");
        for (ipx=0; ipx<8; ipx++) fprintf(illumfil, " %6.0f", cx_cnts[(ipx%7)]);
        fprintf(illumfil, "\n\n");
*/        
   	for (ipy=6; ipy>=0; ipy--) {
           cx_cnts[ipy] = cy_cnts[ipy] = 0.0;
        }
        for (i=0; i<7; i++)
           for (j=0; j<7; j++) {
              c_ideal[i][j] = c_illum[i][j] = 0.0;
              for (k=0; k<7; k++)
                 for (m=0; m<7; m++) 
                    coll_cnts[i][j][k][m] = coll_illum[i][j][k][m] = coll_ideal[i][j][k][m] = 0.0;
           }

        jval++;
        if (jval < 4) goto nextjv;

/*        fclose(gnuxfil);
        fclose(gnuyfil);
        fclose(illumfil);
*/        
        smooth(jemx, accumT);
        
	return;
}

/* *************************************************************** */
         
int sh_update(unsigned short x, unsigned short y, int jemx) 
{
   int i, j;
   double r, xx, yy, dx, dy;
   
   if (y < 185) goto calOK;
   
   dx = (double)(x) - calsrcx[jemx][0];
   dy = (double)(y) - calsrcy[jemx][0];
   
   r = sqrt(dx*dx + dy*dy);
   if (r < calrad[jemx][0]) return(0);
   
   dx = (double)(x) - calsrcx[jemx][1];
   dy = (double)(y) - calsrcy[jemx][1];
   
   r = sqrt(dx*dx + dy*dy);
   if (r < calrad[jemx][1]) return(0);
   
   dx = (double)(x) - calsrcx[jemx][2];
   dy = (double)(y) - calsrcy[jemx][2];
   
   r = sqrt(dx*dx + dy*dy);
   if (r < calrad[jemx][2]) return(0);
   
   dx = (double)(x) - calsrcx[jemx][3];
   dy = (double)(y) - calsrcy[jemx][3];
   
   r = sqrt(dx*dx + dy*dy);
   if (r < calrad[jemx][3]) return(0);
   
calOK: ;

   return(+1);
}

/* *********************************************************** */

int sign(int x)
{
   if (x) return(x/abs(x));
   return(1);
}

/* *********************************************************** */

int sh_xy(double x_rms, double y_rms, float scatt_list[], short scatt_delta[])
{
/*    Calculate the distribution of actual detection points relative to the photon interaction position 
      on a 1x1 mm2 grid. 
      The rms uncertainty in the X and Y position determination (in mm) are given in x_rms and y_rms.
*/      
      int i, j, k, m, n, delta_i[41][41];
      double x_dist[21], y_dist[21], id, t, f, pi_c, q, sum, x, y, xfac, yfac, x_gauss[21], y_gauss[21];
  
      pi_c = 1.0 / sqrt(2.0 * 3.14159);
      
      for (i=0; i<21; i++) {
         id = (double)(2 * i + 1);
         x = 0.5 * id / x_rms;
         f = pi_c * exp(-x * x / 2.0);
         x_gauss[i] = f;
         t = 1.0 / (1.0 + 0.2316419 * x);
         q = f * t * (0.31938153 - t * (.356563782 - t * (1.781477937 - t * (1.821255978 - t * 1.330274429))));
         /* Abramowitz & Stegun, Hdbk of Math. Funct., NBS 1968 */
         x_dist[i] = q; 

         y = 0.5 * id / y_rms;
         f = pi_c * exp(-y * y / 2.0);
         y_gauss[i] = f;
         t = 1.0 / (1.0 + 0.2316419 * y);
         q = f * t * (0.31938153 - t * (.356563782 - t * (1.781477937 - t * (1.821255978 - t * 1.330274429))));
         /* Abramowitz & Stegun, Hdbk of Math. Funct., NBS 1968 */
         y_dist[i] = q; 
      }

      x_dist[20] = x_dist[19];
      y_dist[20] = y_dist[19];
      for (i=19; i>0; i--) {
         x_dist[i] = x_dist[i-1] - x_dist[i];
         y_dist[i] = y_dist[i-1] - y_dist[i];
      }
      x_dist[0] = 1.0 - 2.0 * x_dist[0];
      y_dist[0] = 1.0 - 2.0 * y_dist[0];
      
      sum = -x_dist[0];
      for (i=0; i<21; i++) {
         id = (double)(2 * i + 1);
         x = 0.5 * id / x_rms;
         sum += 2 * x_dist[i];
/*         printf("x: %6.3lf, gauss: %6.4lf, dist: %6.4lf,  sum: %6.4lf\n", x, x_gauss[i], x_dist[i], sum);
*/
      }
      printf("\n");

      sum = -y_dist[0];
      for (i=0; i<21; i++) {
         id = (double)(2 * i + 1);
         y = 0.5 * id / y_rms;
         sum += 2 * y_dist[i];
/*         printf("y: %6.3lf, gauss: %6.4lf, dist: %6.4lf,  sum: %6.4lf\n", y, y_gauss[i], y_dist[i], sum);
*/
      }
      printf("\n");
      
      n = 0;
      for (i=-20; i<=20; i++) {
         k = i + 20;
         xfac = x_dist[(int)(abs(i))];         
         for (j=-20; j<=20; j++) {
            m = j + 20;
            yfac = y_dist[(int)(abs(j))];
            if (xfac * yfac > 0.0001) {
               scatt_list[n] = xfac * yfac;
               scatt_delta[n] = i * det_ydim + j;
               n++;
               if (n >= 200) {printf("Scatt_list length exceeded\n"); return(-1);}
            }
         }
      }

      printf("sh_xy: length of scatter list: %3d\n", n);
      return(n);
}

/* *********************************************************** */

int shapefold(short det[], float comb_list[], short comb_delta[], int comb_len, float det_out[])
{
   int i, j, k;
   double sum1, sum2, loss1;

   for (i=0; i<detdim; i++) det_out[i] = 0.0;
   
   sum1 = loss1 = 0.0;
   for (i=0; i<detdim; i++) {
      if (det[i] > 0.0) {
         sum1 += det[i];
         for (j=0; j<comb_len; j++) {
            k = i + comb_delta[j];
            if ((k < 0) || (k >= detdim)) 
               loss1 += det[i] * comb_list[j];
            else
               det_out[k] += det[i] * comb_list[j];
         }
         if ((i%1000) == 1999) {
            sum2 = 0.0;
            printf("%6d len: %2d: [delta/list]: ", i, comb_len); 
            for (j=0; j<comb_len; j++) {
               printf("%+5d/%6.4f   ", comb_delta[j], comb_list[j]);
               sum2 += comb_list[j];
            }
            printf("  sum: %5.3lf\n", sum2);
         }
      }
   }
   
   sum2 = 0.0;
   for (i=0; i<detdim; i++) sum2 += det_out[i];
/*   for (i=0; i<detdim; i++) det_out[i] /= sum2;
*/
   printf("shapefold: sum_in: %10.4lf,     loss: %10.4lf,    sum_out: %10.4lf\n", sum1, loss1, sum2);
      
   return(0);
}

/* *********************************************************** */

double imagefit(float shdg[], int jemx, int ee, float resid[], int n_source)
{
   int i, j, k, k1, k2, status, rnum[27], activ[25], sxy;
   double mean, var, m0, m1, m2, rms, mm0, mm1, mm2, mmean, mvar, mrms, r, r2, x, y;
   double coeffs[sourcmax * sourcmax], rhs[sourcmax], obs2, fun_k1;
   double coeff2[sourcmax * sourcmax], rhs2[sourcmax], rsum[27];
   
   
   m0 = m1 = m2 = var = 0.0;
   
   for (i=0; i<27; i++) {rnum[i] = 0; rsum[i] = 0.0;}

   for (i=0; i<detdim; i++) {
      if (initmap[i] < 0) continue;	/* all pixels to be rejected are < 0 in initmap[i] */
      m0++;
      m1 += shdg[i];
      m2 += shdg[i] * shdg[i];
      y = (i % 256) - 127.5 + y_shi;
      x = (i / 256) - 127.5 + x_shi;
      r2 = x*x + y*y;
      r = sqrt(r2);
      k = r / 5;
      if (k > 26) k = 26;
      rnum[k]++;
      rsum[k] += shdg[i];
   }
   mean = m1 / m0;
   var = m2/m0 - mean * mean;
   rms = sqrt(var);
   activ[0] = m0;
   activ[1] = m0;
   activ[2] = m0;

   
   if (n_source <= 3) {
      printf("\nJEM-X%1d, E=%1d, events: %10.0lf, active pixels: %7.0lf, counts/pixel: %7.1lf, rms: %7.2lf\n",
         jemx+1, ee, m1, m0, mean, rms);
      
/*      printf("Radial distribution: \n");
      for (i=0; i<25; i++) printf(" %4d", rnum[i]);
      printf("\n");
      for (i=0; i<25; i++) printf(" %4d", (int)(rsum[i]/rnum[i]));
      printf("\n");
*/
   }
   
   /* collect data */
   
   /* n_source = 2   constant + diffuse background in matching energy band */

   for (i=0; i<n_source; i++) {
      rhs[i] = 0.0;
      for (j=0; j<n_source; j++) coeffs[i*n_source+j] = 0.0;
   }
   
   for (i=0; i<detdim; i++) {
      if (initmap[i] < 0) continue;	/* all pixels to be rejected are <0 in initmap */
      for (k1=0; k1<n_source; k1++) {
         fun_k1 = fun(k1, i, ee, activ[k1]);
         rhs[k1] += fun_k1 * shdg[i];
         for (k2=k1; k2<n_source; k2++) coeffs[k1 + n_source*k2] += fun_k1 * fun(k2, i, ee, activ[k2]);
      }
   }
   for (k1=0; k1<n_source; k1++) {
      rhs2[k1] = rhs[k1];
      for (k2=k1; k2<n_source; k2++) {
         coeffs[k2+n_source*k1] = coeff2[k1+n_source*k2] = coeff2[k2+n_source*k1] = coeffs[k1+n_source*k2];
      }
   }
      
   for (k1=0; k1<n_source; k1++) {
      for (k2=0; k2<n_source; k2++) {
         printf(" %8.1lf", coeffs[k1*n_source+k2]);
      }
      printf("   %8.1lf\n", rhs[k1]);
   }
   
   /* solve the equation system */
   status = gaussj_nr(coeffs, n_source, rhs, 1);
   if (status < 0) {printf("Singular matrix: %2d\n", status); return(status);}
   
   for (k1=0; k1<n_source; k1++) solut[k1] = rhs[k1];

   for (i=0; i<detdim; i++) {
      resid[i] = 0.0;
      if (initmap[i] < 0) continue;	/* all pixels to be rejected are <0 in initmap */
      resid[i] = shdg[i];      
      for (k1=0; k1<n_source; k1++) {
         resid[i] -= solut[k1] * fun(k1, i, ee, activ[k1]);
      }
   }
   
   printf("\n");
   for (k1=0; k1<n_source; k1++) {
      for (k2=0; k2<n_source; k2++) {
         printf(" %8.1lf", coeffs[k1*n_source+k2]);
      }
      printf("   %8.1lf\n", solut[k1]);
   }
   
   mm0 = mm1 = mm2 = mvar = 0.0;

   for (i=0; i<27; i++) {rnum[i] = 0; rsum[i] = 0.0;}
   
   for (i=0; i<detdim; i++) {
      if (initmap[i] < 0) continue;	/* all pixels to be rejected are <0 in initmap[i] */
      mm0++;
      mm1 += resid[i];
      mm2 += resid[i] * resid[i];
      y = (i % 256) - 127.5 + y_shi;
      x = (i / 256) - 127.5 + x_shi;
      r2 = x*x + y*y;
      r = sqrt(r2);
      k = r / 5;
      if (k > 26) k = 26;
      rnum[k]++;
      rsum[k] += resid[i];
   }
   mmean = mm1 / mm0;
   mvar = mm2/mm0 - mmean * mmean;

   if (mvar < 0.0) {printf("Negative variance after fit: %10.1lf\n", mvar); return(-1);}
   mrms = sqrt(mvar);   

   printf("JEM-X%1d, E=%1d, after fitting with %2d components: ", jemx+1, ee, n_source);
   printf(" Counts/pixel: %10.5lf, rms: %7.2lf\n", mmean, mrms);

   if (n_source <= 3) {
/*      printf("Radial distribution: \n");
      for (i=0; i<25; i++) printf(" %4d", rnum[i]);
      printf("\n");
      for (i=0; i<25; i++) printf(" %4d", (int)(rsum[i]/rnum[i]));
      printf("\n\n");
*/
   }
   
   return(mrms);
}

/* **************************************************************** */

float fun(int k, int ii, int ee, int active)
{
   static int src_list[sourcmax] = {0, 0};
   int i, j, n, sxy, length, bytes, det_index, word, bitval;
   unsigned char irosbytes[16000];
   static unsigned src_bits[sourcmax][2048], cnt=0;
   static double norm[sourcmax];
   double val;
   
   if (k < 0) for (i=0; i<sourcmax; i++) src_list[i] = 0;
   
   if (initmap[ii] > 0) {
      if (k == 0) return(1.0 / (float)(active));
      if (k == 1) return(fun_shdgr[ii+yshift]);
   }

   return(0.0);

}

/* **************************************************************** */

int gaussj_nr(double a[], int n, double b[], int m)

/* -------------------------------------------------------------

docbegin:

*  Name of source file:
*         /home/nl/jemx/ipfread/jemx1-2/sky/gaussj_nr.c

*  Version of source file:      0.00
*  Parent component:            gaussj_nr from NJW 
*  Programmer:                  Niels Lund (modification to avoid DAL/PIL/RIL dependencies)
*  Affiliation:                 Danish Space Research Institute
*  Purpose:                     Fit of shadowgram components for Iterative Removal of Sources
*  Origin date:                 021220
*  Update history:
*  Type of element:             Function
*  Part of package:             minisky.c
*  Required functions		SWAP
Various remarks:

Gauss-Jordan matrix inversion from "Numerical Recipes"
adapted to follow the storage convention:		    a(i,j) = a[j*Ni + i]

-------------------------------------------------------------- */
{
   int icol, irow, i, j, k, l, ll;
   int indxc[20], indxr[20], ipiv[20];
   double big, dum, pivinv;

   for (j=0;j<n;j++) ipiv[j]=0;
   for (i=0;i<n;i++) {
      big=0.0;
      for (j=0;j<n;j++)
         if (ipiv[j] != 1)
            for (k=0;k<n;k++) {
               if (ipiv[k] == 0) {
                  if (fabs(a[k*n+j]) >= big) {
                     big=fabs(a[k*n+j]);
                     irow=j;
                     icol=k;
                  }
               } else if (ipiv[k] > 1) return(-1);	/* singular matrix 1 */
            }
      ++(ipiv[icol]);
      
      if (irow != icol) {
         for (l=0;l<n;l++) SWAP(a[l*n + irow],a[l*n + icol]);
         for (l=0;l<m;l++) SWAP(b[l*n + irow],b[l*n + icol]);
      }
      indxr[i]=irow;
      indxc[i]=icol;
      if (a[icol*n + icol] == 0.0) return(-2);	/* singular matrix 2 */
      
      pivinv=1.0/a[icol*n + icol];
      a[icol*n + icol]=1.0;
      for (l=0;l<n;l++) a[l*n + icol] *= pivinv;
      for (l=0;l<m;l++) b[l*n + icol] *= pivinv;
      for (ll=0;ll<n;ll++)
         if (ll != icol) {
            dum=a[icol*n + ll];
            a[icol*n + ll]=0.0;
            for (l=0;l<n;l++) a[l*n + ll] -= a[l*n + icol]*dum;
            for (l=0;l<m;l++) b[l*n + ll] -= b[l*n + icol]*dum;
         }
   }
   for (l=n-1;l>=0;l--) {
      if (indxr[l] != indxc[l])
         for (k=0;k<n;k++)
            SWAP(a[indxr[l]*n + k],a[indxc[l]*n + k]);
   }
   return(0);
}


/* *********************************************************** */

int pixelfold(short det[], int pixel_st[], unsigned char pixel_list[], float det_out[])
{
   int i, j, k, m, n, len, index, delta, val, valnum;
   double sum1, sum2, loss1, valsum, vals;

   for (i=0; i<detdim; i++) det_out[i] = 0.0;
   valsum = vals = 0.0;
   valnum = 0;
   
   sum1 = loss1 = 0.0;
   for (i=1; i<detdim; i++) {
      if ((det[i] <= 0) || (initmap[i] < 0)) continue;
      sum1 += det[i];
      vals = 0.0;
      index = pixel_st[i];
      len = pixel_list[index++];
      for (j=0; j<len; j++) {
         delta = pixel_list[index++];
         val = pixel_list[index++];
         vals += val;
         k = ((delta & 15) - 8); 
         m = ((delta & 240) - 128) * 16;
         n = i + k + m;
         if ((n < 0) || (n >= detdim)) 
            loss1 += val * (float)(det[i]);
         else
            det_out[n] += val * (float)(det[i]);         
      }
      valsum += vals / 255.0;
      valnum++;
      if ((vals > 260.0) || (vals < 250.0)) {
         index = pixel_st[i];
         len = pixel_list[index++];
         printf("%6d pixel: len: %1d, [delta/delta'/val]: ", i, len);
         n = 0;
         for (j=0; j<len; j++) {
            delta = pixel_list[index++];
            val = pixel_list[index++];
            n += val;
            k = ((delta & 15) - 8); 
            m = ((delta & 240) - 128) * 16;
            m += k;
            printf(" %2x/%+4d/%3d  ", delta, m, val);
         }
         printf(" sum: %3d\n", n);
         if ((vals > 260.0) || (vals < 250.0)) exit(0);
      }
   }
   
   sum2 = 0.0;
   for (i=0; i<detdim; i++) {
      if (initmap[i] < 0) det_out[i] = 0;
      det_out[i] /= 255.0;
      sum2 += det_out[i];
   }
   loss1 /= 255.0;
/*   for (i=0; i<detdim; i++) det_out[i] /= sum2;
*/
   printf("pixelfold: sum_in: %10.4lf,     loss: %10.4lf,    sum_out: %10.4lf, average valsum: %6.4lf\n", 
      sum1, loss1, sum2, valsum/valnum);
      
   return(0);
}

/* ******************************************************************************** */

void smooth(int jemx, double accumT)
/* generate smoothed shdgr, summing original shdgr over 3 pixels in the radial direction 
   and 3 pixels in the local tangential direction
*/
{
   int i, j, k, m, n, mj, mi, ii, jj, ee=0, pmin=0, cnt, det_index, det_index2;
   double fac0 = 9.0;	/* There will be 9 times the true number of counts in the images  */
   int rsmooth = 3;	/* must be an odd number !!! */
   int tsmooth = 3;	/* must be an odd number !!! */
   int rsm, tsm, num_tot[4];
   float smo_shdgr[256][256];
   char filnavn[100]; 
   double kistep, kjstep, rrdet, xrdet, yrdet, ki, kj, c_fac, printfac, smo_sum;
   char charm[8] = {'0', 'S', 'Z', 'U', '*', '-', '-', '-'};
   FILE *soft_bin, *sm_fil, *soft_txt;
        
   if (soft_tot[25] == 1) sprintf(filnavn, "smooth%1d+anod.bin", jemx+1);
   else  sprintf(filnavn, "smooth%1d-anod.bin", jemx+1);
   soft_bin = fopen(filnavn, "wb");
   
   if (soft_tot[25] == 1) sprintf(filnavn, "smooth%1d+anod.txt", jemx+1);
   else  sprintf(filnavn, "smooth%1d-anod.txt", jemx+1);
   soft_txt = fopen(filnavn, "wt");
   
   rsm = rsmooth / 2;
   tsm = tsmooth / 2;
   
next_e:
   k = 0;
   for (i=0; i<256; i++) 
      for (j=0; j<256; j++) {
         det_index = 256 * i + j;
         if (initmap[det_index] > 0) {
            smo_shdgr[i][j] = 0.0;
            k++;
         } else smo_shdgr[i][j] = initmap[det_index];
      }
   printf("Initmap accepted positions: %5d\n", k);
   
   for (i=0; i<256; i++) {
      xrdet = (double)(i) - 127.5 + x_shi;
      for (j=0; j<256; j++) {
         det_index = 256 * i + j;
         if (smo_shdgr[i][j] < 0.0) continue;

         yrdet = (double)(j) - 127.5 + y_shi;
         rrdet = sqrt(xrdet*xrdet + yrdet*yrdet);
                  
         if ((xrdet == 0.0) || (yrdet == 0.0)) {
            if (xrdet == 0.0) {
               ki = mj = 1; 
               kj = mi = 0;
            } else {
               ki = mj = 0; 
               kj = mi = 1;
            }
         } else {
            kistep = yrdet / rrdet;
            kjstep = -xrdet / rrdet;
            if (fabs(kjstep) > fabs(kistep)) {
               kj = (int)(1.1 * kjstep / fabs(kjstep));
               ki = kistep / fabs(kjstep);
               mi = 1;
               mj = 0;
            } else {
               kj = kjstep / fabs(kistep);
               ki = (int)(1.1 * kistep / fabs(kistep));
               mj = 1;
               mi = 0;
            }
         }
         
         cnt = 0;
         for (m=-rsm; m<=rsm; m++) {
            for (k=-tsm; k<=tsm; k++) {
               ii = i + (int)(k*ki + .5) + m*mi;
               jj = j + (int)(k*kj + .5) + m*mj;
               if (ii < 0) ii = 0;
               if (ii > 255) ii = 255;
               if (jj < 0) jj = 0;
               if (jj > 255) jj = 255;
               det_index2 = 256 * ii + jj;
               if ((soft_shdgr[ee][det_index2] > 0.0) && (smo_shdgr[i][j] >= 0.0)) {
                  smo_shdgr[i][j] += soft_shdgr[ee][det_index2];
                  cnt++;
               }
            }
         }
         if (cnt > 0) {
            c_fac = (double)(rsmooth*tsmooth) / (double)(cnt);  
            smo_shdgr[i][j] = smo_shdgr[i][j] / c_fac + 0.5; 
         }
      }
   }
        
   sprintf(filnavn, "smo_maps%1d_%1d.dat", jemx+1, ee);
/*   sm_fil = fopen(filnavn, "wt");
   if (sm_fil == NULL) {printf("Could not open %s file\n", filnavn); return;}
*/
   soft_tot[ee] = 0.0;
   num_tot[ee] = 0;
   for (i=0; i<256; i++) {
      ii = 255 - i;
      for (j=0; j<256; j++) {
         if (smo_shdgr[j][ii] >= 0.0) {
            soft_tot[ee] += smo_shdgr[j][ii];
            num_tot[ee]++;
         }
      }
   }
   printf("Before [sensi]. Softsum for E=%1d: %14.1lf, contributing pixels: %6d\n", 
      ee, soft_tot[ee], num_tot[ee]);

   soft_tot[26+ee] = sensi(ee, num_tot[ee], smo_shdgr);
   
   soft_tot[ee] = num_tot[ee] = 0;
   printfac = limit[ee][m_max] / 16.0 + 0.5;

   soft_tot[ee] = 0.0;
   num_tot[ee] = 0;
   for (i=0; i<256; i++) {
      ii = 255 - i;
      for (j=0; j<256; j++) {
         det_index = 256 * j + ii;
         if (smo_shdgr[j][ii] >= 0.0) {
            soft_tot[ee] += smo_shdgr[j][ii];
            num_tot[ee]++;
         }
/*         if (smo_shdgr[j][ii] > 0) fprintf(sm_fil, "%1x", (int)(smo_shdgr[j][ii] / printfac));
         else if ((smo_shdgr[j][ii] == 0) && (initmap[det_index] == -1)) fprintf(sm_fil, " ");
              else fprintf(sm_fil, "%1c", charm[(int)(-smo_shdgr[j][ii]-1)]);
*/
      }
/*      fprintf(sm_fil, "\n");
*/
   }
/*   fprintf(sm_fil, "\n\n");
   fclose(sm_fil);
*/   
   printf("After [sensi]. Softsum for E=%1d: %14.1lf, contributing pixels: %6d %6.0lf ", 
      ee, soft_tot[ee], num_tot[ee], soft_tot[26+ee]);
   
   for (i=0; i<256; i++) {
      for (j=0; j<256; j++) {
         if (smo_shdgr[i][j] >= 0.0) smo_shdgr[i][j] /= soft_tot[ee];
/* 	fprintf(soft_txt, "%+9.7f\n", smo_shdgr[i][j]);
*/
      }
   }

   fwrite(smo_shdgr, 4, 65536, soft_bin);

   smo_sum = 0.0;
   for (i=0; i<256; i++) {
      ii = 255 - i;
      for (j=0; j<256; j++) {
         if (smo_shdgr[j][ii] > 0.0) smo_sum += smo_shdgr[j][ii];
      }
   }
   printf("  Final sum of all positive pixels: %6.4lf\n", smo_sum);

   ee++;
   if (ee < 4) goto next_e;

   soft_tot[4] = accumT;
   j = 5;
   for (i=0; i<4; i++) {
      soft_tot[j++] = calsrcx[jemx][i];	/* cal source positions & elimination radii */
      soft_tot[j++] = calsrcy[jemx][i];
      soft_tot[j++] = calrad[jemx][i];
   }
   soft_tot[17] = pha_sel[0];	/* pulseheight selection boundaries */
   soft_tot[18] = pha_sel[1];
   soft_tot[19] = pha_sel[2];
   soft_tot[20] = pha_sel[3];
   soft_tot[21] = pha_sel[4];
   soft_tot[22] = fac0;
   soft_tot[23] = ref_tot[23];
   soft_tot[24] = ref_tot[24];
   
   fwrite(soft_tot, 8, 30, soft_bin);
   fclose(soft_bin);
      
   for (j=0; j<30; j++) fprintf(soft_txt, "%+14.7lf\n", soft_tot[j]);
   fclose(soft_txt);
   
   return;
}

/* ******************************************************************************** */

int sensi(int ee, int num, float ss[256][256])
{
   int i, j, k, m, k_max, index, ll, ll_list[600], m_start, m_end, mm, det_index;
   float s[65536], lim;
   double signal, noise, SN, sn_max, sum=0.0;
   
   signal = noise = SN = 0.0;
   
   m_start = 0;
   for (i=0; i<256; i++) {
      for (j=0; j<256; j++) {
         det_index = 256*i + j;
         if (initmap[det_index] > 0) {
            s[det_index] = ss[i][j];
            sum += s[det_index];
            m_start++;
         } else s[det_index] = -10000 + initmap[det_index];
      }
   }
   k_max = m_start / 100;
   
   printf("Sensi: input sum: %10.1lf, contributing pixels: %6d, k_max: %6d\n", sum, m_start, k_max);
   
   sn_max = 0.0;
   for (m=0; m<100; m++) {
      for (k=0; k<k_max; k++) {
         lim = 1000000.0;
         ll = -1;
         for (index=0; index<detdim; index++) {
            if ((s[index] > 0) && (s[index] < lim)) {
               lim = s[index];
               ll = index;
            }
         }
         if (ll >= 0) {
            s[ll] = -s[ll];
            ll_list[k] = ll;
         } else break;
      }
      
      if (lim < 1000000.0) limit[ee][m] = lim;
      else limit[ee][m] = limit[ee][m-1];
      if (m < 50) continue;
      if (m == 50) {
         for (mm=0; mm<50; mm++) {
            signal = mm+1;
            noise += lim;
            SN = signal / sqrt(noise);
            snratio[ee][mm] = SN;
         }
      }
      
      if (lim < 1000000.0) {
         signal = m+1;
         noise += lim;
      }
      SN = signal / sqrt(noise);
      if (SN > sn_max) {
         sn_max = SN;
         m_max = m;
      } else for (j=0; j<k; j++) s[ll_list[j]] = -10004;
      
      snratio[ee][m] = SN;
      printf("%3d  %6.1f  %10.0lf  %10.5lf\n", m+1, limit[ee][m], noise, SN);      
   }
   
   m_end = 0;
   index = 0;
   for (i=0; i<256; i++) {
      for (j=0; j<256; j++) {
         if (s[index+j] > -10000) {
            ss[i][j] = -s[index+j];
            m_end++;
         } else ss[i][j] = s[index+j] + 10000;
      }
      index += 256;
   }
   
   printf("area_start: %5d,  area_end: %5d  m_max: %3d\n", m_start, m_end, m_max);
   
   return(m_end);
}
/* ******************************************************************************** */


